import json
import logging
import os
from typing import Any, Callable

import yaml
from jinja2 import Template

MODEL_JSON_FILE = "/model.json"


class BaseModel(object):

    def __init__(self, repo_manager):
        self.repo_manager = repo_manager

    @staticmethod
    def get_model_ids(model_path: str, rtn_val=False) -> dict:
        catalog = {}
        index = 0
        model_ids = "Use one of the following models\n"

        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + MODEL_JSON_FILE):
                with open(model_path + model_folder + MODEL_JSON_FILE, "r") as f:
                    model_definition = json.load(f)
                    catalog[index] = model_definition
                    index += 1

        for key in catalog:
            model_ids += "{1}: {0}\n".format(catalog[key]["id"], catalog[key]["name"])

        if rtn_val:
            return catalog

        raise ValueError(model_ids)

    @staticmethod
    def get_model_folders(model_path: str, rtn_val=False) -> dict:
        catalog = {}
        index = 0
        model_ids = "Use one of the following models\n"
        model_folder = None

        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + MODEL_JSON_FILE):
                with open(model_path + model_folder + MODEL_JSON_FILE, "r") as f:
                    model_definition = json.load(f)
                    catalog[index] = model_definition
                    index += 1

        for key in catalog:
            model_ids += "{1}: {0}\n".format(model_folder, catalog[key]["name"])

        if rtn_val:
            return catalog

        raise ValueError(model_ids)

    @staticmethod
    def get_model_folder(model_path: str, model_id: str) -> str:
        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + MODEL_JSON_FILE):
                with open(model_path + model_folder + MODEL_JSON_FILE, "r") as f:
                    model_definition = json.load(f)
                    if model_definition["id"] == model_id:
                        return model_folder

        raise ValueError("Could not find model path for model_id: {}".format(model_id))

    def __get_model_varargs(self, model_id: str) -> dict[str, str]:  # noqa #NOSONAR
        return {
            "model_id": model_id,
            "model_version": "cli",
            "model_table": "vmo_models_cli",
            "project_id": self.__get_project_id(),
            "job_id": "cli",
        }

    def __get_project_id(self) -> str:
        self.repo_manager.repo_config_rename(self.repo_manager.base_path)
        path = os.path.join(self.repo_manager.base_path, ".tmo/config.yaml")
        with open(path, "r") as handle:
            return yaml.safe_load(handle)["project_id"]

    @staticmethod
    def __template_sql_script(filename: str, jinja_ctx: Any) -> str:
        with open(filename) as f:
            template = Template(f.read(), autoescape=True)

        return template.render(jinja_ctx)

    def __execute_sql_script(self, filename: str, jinja_ctx: Any):  # noqa #NOSONAR
        from ..util.connections import execute_sql

        script = self.__template_sql_script(filename, jinja_ctx)

        stms = script.split(";")

        for stm in stms:
            stm = stm.strip()
            if stm:
                logging.info("Executing statement: {}".format(stm))

                try:
                    execute_sql(stm)
                except Exception as e:
                    if stm.startswith("DROP"):
                        logging.warning("Ignoring DROP statement exception")
                    else:
                        raise e

    @staticmethod
    def __get_engine(  # noqa #NOSONAR
        model_definition: dict[str, Any], mode: str
    ) -> dict:
        if "automation" in model_definition:

            if (
                mode in model_definition["automation"]
                and "engine" in model_definition["automation"][mode]
            ):
                return model_definition["automation"][mode]["engine"]

            # legacy
            if "trainingEngine" in model_definition["automation"]:
                return model_definition["automation"]["trainingEngine"]

        return model_definition["language"]

    def __run_r_model(  # noqa #NOSONAR
        self, model_id: str, model_dir: str, data_conf: dict, mode: str
    ):
        import tempfile
        import subprocess

        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(json.dumps(data_conf).encode())

        pkg_path, _ = os.path.split(__file__)
        cmd = (
            f"{pkg_path}/run_model.R"
            f" {model_id} {self.__get_project_id()} {mode} {fp.name} {model_dir}"
        )
        subprocess.check_call(cmd, shell=True)

    def _execute_model(  # noqa
        self,
        model_definitions_path: str,
        model_id: str,
        rendered_dataset: dict,
        automation_mode: str,
        execute_callback: Callable,
    ):
        """
        Método común para ejecutar modelos (score, evaluate, etc).

        Args:
            model_definitions_path: Ruta al directorio de definiciones de modelos
            model_id: ID del modelo
            rendered_dataset: Dataset renderizado
            automation_mode: Modo de automatización (DEPLOY, EVALUATE, etc)
            execute_callback: Función callback que ejecuta la lógica específica.
                             Recibe (engine, context, model_dir, model_definition,
                                     model_conf, rendered_dataset, cli_model_kwargs)
        """
        from tmo.context.model_context import ModelContext, DatasetInfo
        from tmo.types.job import JobRunnerEngine

        model_dir = model_definitions_path + BaseModel.get_model_folder(
            model_definitions_path, model_id
        )

        with open(f"{model_dir}/model.json", "r") as f:
            model_definition = json.load(f)

        with open(f"{model_dir}/config.json", "r") as f:
            model_conf = json.load(f)

        logging.getLogger(__name__).info("Loading and executing model code")

        cli_model_kwargs = self.__get_model_varargs(model_id)

        context = ModelContext(
            dataset_info=DatasetInfo.from_dict(rendered_dataset),
            hyperparams=model_conf["hyperParameters"],
            artifact_input_path="artifacts/input",
            artifact_output_path="artifacts/output",
            **cli_model_kwargs,
        )

        engine = JobRunnerEngine(self.__get_engine(model_definition, automation_mode))

        execute_callback(
            engine=engine,
            context=context,
            model_dir=model_dir,
            model_definition=model_definition,
            model_conf=model_conf,
            rendered_dataset=rendered_dataset,
            cli_model_kwargs=cli_model_kwargs,
        )



from tmo.stats.stats import *
from tmo.stats.stats_util import *
from tmo.stats.stats_util import (
    _NpEncoder,
    _compute_continuous_edges,
    _convert_all_edges_to_val_str,
    _fill_missing_bins,
    _process_categorical_var,
    _process_continuous_var,
    _parse_scoring_stats,
    _capture_stats,
    _histogram_wrapper,
    _val_parameter_length_too_long,
    _round_to_td_float,
)



from __future__ import absolute_import

from tmo.api.base_api import BaseApi


class MessageApi(BaseApi):

    path = "/api/message"
    type = "MESSAGE"

    def send_job_event(self, event: dict):
        """
        send a job event

        Parameters:
           event (dict): event to send

        Returns:
            (dict): event
        """

        return self._send_event("jobevents", event)

    def send_progress_event(self, event: dict):
        """
        send a progress event

        Parameters:
           event (dict): event to send

        Returns:
            (dict): event
        """

        return self._send_event("jobprogress", event)

    def _send_event(self, topic, event):

        header_params = {"Content-Type": self.json_type}
        query_params = {"type": "topic"}

        return self.tmo_client.post_request(
            f"{self.path}/{topic}", header_params, query_params, event
        )



import importlib
import json
import logging
import os
import shutil
import sys

from tmo.cli.base_model import BaseModel
from tmo.cli.model_utils import cleanup_artifacts, ARTIFACTS_PATH
from tmo.context.model_context import ModelContext, DatasetInfo
from tmo.types.job import JobRunnerEngine, JobAutomationMode


class EvaluateModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def evaluate_model_local(
        self, model_id: str, rendered_dataset: dict, base_path: str
    ):

        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = f"{base_path}model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(f"model directory {model_definitions_path} does not exist")

        model_artefacts_path = f".artefacts/{model_id}"
        model_evaluation_path = f"{model_artefacts_path}/evaluation"

        if not os.path.exists(f"{model_artefacts_path}/output"):
            raise ValueError("You must run training before trying to run evaluation.")

        if os.path.exists(model_evaluation_path):
            self.logger.debug(
                f"Cleaning local model evaluation path: {model_evaluation_path}"
            )
            shutil.rmtree(model_evaluation_path)

        os.makedirs(f"{model_evaluation_path}/output")

        try:
            cleanup_artifacts()

            os.makedirs(ARTIFACTS_PATH)

            os.symlink(
                f"../{model_artefacts_path}/output/",
                f"./{ARTIFACTS_PATH}/input",
                target_is_directory=True,
            )
            os.symlink(
                f"../{model_artefacts_path}/evaluation/output/",
                f"./{ARTIFACTS_PATH}/output",
                target_is_directory=True,
            )

            def execute_evaluate(
                engine,
                context,
                model_dir,
                model_definition,
                model_conf,
                rendered_dataset,
                cli_model_kwargs,
            ):
                if engine == JobRunnerEngine.PYTHON:
                    self.__evaluate_python(
                        context,
                        model_dir,
                        rendered_dataset,
                        model_conf,
                        **cli_model_kwargs,
                    )

                elif engine == JobRunnerEngine.SQL:
                    self.__evaluate_sql(
                        context,
                        model_dir,
                        rendered_dataset,
                        model_conf,
                        **cli_model_kwargs,
                    )

                elif engine == JobRunnerEngine.R:
                    self._BaseModel__run_r_model(  # noqa
                        model_id, model_dir, rendered_dataset, "evaluate"
                    )

                else:
                    raise ValueError(
                        f"Unsupported language: {model_definition['language']}"
                    )

            self._execute_model(
                model_definitions_path=model_definitions_path,
                model_id=model_id,
                rendered_dataset=rendered_dataset,
                automation_mode=JobAutomationMode.EVALUATE,
                execute_callback=execute_evaluate,
            )

            if os.path.exists(f"{model_artefacts_path}/evaluation/output/"):
                self.logger.info(
                    "Artefacts can be found in:"
                    f" {model_artefacts_path}/evaluation/output/"
                )
            else:
                self.logger.info(f"Artefacts can be found in: {model_artefacts_path}")

            if os.path.exists(f"{model_artefacts_path}/evaluation/output/metrics.json"):
                self.logger.info(
                    "Evaluation metrics can be found in:"
                    f" {model_artefacts_path}/evaluation/output/metrics.json"
                )

            if os.path.exists(f"{model_artefacts_path}/evaluation.json"):
                self.logger.info(
                    "Evaluation metrics can be found in:"
                    f" {model_artefacts_path}/evaluation.json"
                )

            cleanup_artifacts()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            cleanup_artifacts()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                f"pip install -r {model_dir}/model_modules/requirements.txt"
            )
            raise
        except:
            cleanup_artifacts()
            self.logger.exception("Exception running model code")
            raise

    @staticmethod
    def __evaluate_python(context, model_dir, rendered_dataset, model_conf, **kwargs):
        sys.path.append(model_dir)
        if os.path.isfile(f"{model_dir}/model_modules/evaluation.py"):
            evaluation = importlib.import_module(".evaluation", package="model_modules")
        else:
            logging.debug("No evaluation.py found. Using scoring.py -> evaluate")
            evaluation = importlib.import_module(".scoring", package="model_modules")

        evaluation.evaluate(
            context=context,
            data_conf=rendered_dataset,
            model_conf=model_conf,
            **kwargs,
        )

    def __evaluate_sql(self, context, model_dir, data_conf, model_conf, **kwargs):
        from tmo.util import tmo_create_context
        from teradataml import remove_context, DataFrame

        self.logger.info("Starting evaluation...")

        tmo_create_context()  # noqa

        sql_file = f"{model_dir}/model_modules/evaluation.sql"
        jinja_ctx = {
            "context": context,
            "data_conf": data_conf,
            "model_conf": model_conf,
            "model_table": kwargs.get("model_table"),
            "model_version": kwargs.get("model_version"),
            "model_id": kwargs.get("model_id"),
        }

        self._BaseModel__execute_sql_script(sql_file, jinja_ctx)  # noqa

        self.logger.info("Finished evaluation")

        stats = DataFrame(data_conf["metrics_table"]).to_pandas(all_rows=True)
        metrics = dict(zip(stats.key, stats.value))

        with open("models/evaluation.json", "w+") as f:
            json.dump(metrics, f)

        self.logger.info("Saved metrics")

        remove_context()



from enum import Enum

from tmo.decorators.functional import functional


class DataType(Enum):
    FLOAT = "float"
    INTEGER = "integer"


class TypeEnum(Enum):
    ENTITY = "entity"
    FEATURE = "feature"
    TARGET = "target"


class CatalogType(Enum):
    VANTAGE = "VANTAGE"
    NO_CATALOG = "NO_CATALOG"


class CatalogBodyType(Enum):
    VANTAGE = "CatalogBody"
    NO_CATALOG = "NoCatalogBody"


@functional
class Variable(object):
    name: str = None  # NOSONAR
    data_type: DataType = None  # NOSONAR
    type: TypeEnum = None  # NOSONAR
    selected: str = None  # NOSONAR
    entity_id: str = None  # NOSONAR


@functional
class FeaturesEntityTargets(object):
    entity: str = None  # NOSONAR
    sql: str = None  # NOSONAR
    variables: list[Variable] = []

    def set_columns(self, columns: list[Variable]) -> object:
        join = self.variables + columns
        self.variables = join
        return self

    def add_column(self, variable: Variable) -> object:
        v = [variable]
        join = self.variables + v
        self.variables = join
        return self


@functional
class Predictions(object):
    database: str = None  # NOSONAR
    entity_sql: str = None  # NOSONAR
    table: str = None  # NOSONAR


@functional
class Metadata(object):
    entity_and_targets: FeaturesEntityTargets = None  # NOSONAR
    features: FeaturesEntityTargets = None  # NOSONAR
    predictions: Predictions = None  # NOSONAR
    type: CatalogBodyType = None  # NOSONAR


@functional
class FeatureMetadata(object):
    database: str = None  # NOSONAR
    table: str = None  # NOSONAR



from __future__ import absolute_import

import logging
import uuid
import warnings
from typing import Any, Optional, Union

import pandas as pd
from teradataml import DataFrame, execute_sql

from tmo.api.iterator_base_api import IteratorBaseApi
from tmo.types.dataset_metadata import (
    CatalogBodyType,
    CatalogType,
    DataType,
    FeatureMetadata,
    FeaturesEntityTargets,
    Metadata,
    Predictions,
    TypeEnum,
    Variable,
)
from tmo.types.dataset_template import DatasetTemplate
from tmo.types.exceptions import EntityCreationError

logger = logging.getLogger(__name__)


class DatasetTemplateApi(IteratorBaseApi):
    path = "/api/datasetTemplates"
    type = "DATASET_TEMPLATE"

    def create(
        self,
        name: str = "SDK Dataset Template",
        columns: dict[str, list[str]] = None,
        dataframe: Optional[DataFrame | pd.DataFrame] = None,
        database: Optional[str] = "TD_MODELOPS",
        tables: Optional[dict] = None,
        sql: Optional[dict] = None,
        description: Optional[str] = "SDK dataset template",
        catalog_type: Optional[CatalogType] = CatalogType.VANTAGE,
    ) -> DatasetTemplate:
        """
        Initialize a DatasetTemplate object.

        Parameters:
            name (str): The name of the dataset template
            description (str): A brief description of the dataset template. (Default is "VANTAGE dataset template")
            columns (list[str]): A list containing the target and entity columns in the dataset template
            dataframe (DataFrame): The dataframe to be used for creating the dataset template
            database (str, optional): The name of the database. (Default is "td_modelops")
            catalog_type (CatalogType, optional): The type of catalog to use. (Default is "VANTAGE")
            tables (dict, optional): A dictionary containing the table names for feature metadata, data, and predictions. If None, the tables will be automatically created using the dataset template name as a base.
                - data: Contains entity, features and targets.
                - features: Contains features.  If not present data table will be used.
                - entityTarget: Contains entity and targets. If not present data table will be used.
                - featureMetadata: Stores metadata about the features.
                - predictions: Holds prediction results generated during model evaluation.
            sql (dict, optional): A dictionary containing the SQL queries for features, entity and target, and predictions. If None, the SQL queries will be automatically generated based on the provided dataframe.
                Required sql queries:
                - features: SQL query for extracting features
                - target_entity: SQL query for extracting entity and targets
                - predictions: SQL query for generating predictions

        Returns:
            dict: The created dataset template.

        Example:
            ```python
            from tmo import TmoClient
            from teradataml import DataFrame

            con = create_context(host="10.15.126.184",username="admin",password="admin",database="td_modelops")

            data = DataFrame.from_table("PIMA")

            vmoClient = TmoClient()

            dataset_template = (
                vmoClient
                .dataset_templates()
                .create(
                    name="New Dataset Template",
                    columns={
                        "entity": ["PatientId"],
                        "targets": ["HasDiabetes"],
                    },
                    dataframe=data,
                    database="my_database",
                    tables={
                        "data": "pima_patient_data",
                        "featureMetadata": "pima_statistics_metadata",
                        "predictions": "pima_predictions",
                    },
                    sql={
                        "features": "SELECT * FROM pima_patient_data",
                        "target_entity": "SELECT * FROM pima_patient_diagnoses",
                        "predictions": "SELECT * FROM pima_patient_data F WHERE F.patientId MOD 5 = 0",
                    }
                )
            )
            ```
        """

        sql = sql or {}
        tables = tables or {}
        columns = columns or {}

        if (
            columns is None
            or not isinstance(columns, dict)
            or "entityColumns" not in columns
            or "targetColumns" not in columns
        ):
            raise ValueError(
                "Columns must be a dictionary with 'entityColumns' and 'targetColumns'"
                " keys."
            )

        if dataframe is None or not isinstance(dataframe, DataFrame):
            raise ValueError("Dataframe must be a valid teradataml DataFrame object.")

        entity_columns = columns["entityColumns"]
        target_columns = columns["targetColumns"]

        dataset_template = DatasetTemplate()
        dataset_template.name = name
        dataset_template.description = description
        dataset_template.catalog_type = catalog_type
        dataset_template.entity = entity_columns[0]
        dataset_template.target = target_columns

        data_table = tables.get(
            "data", f"{dataset_template.name.replace(' ', '_')}_data"
        )
        entity_target_table = tables.get("entityTarget", data_table)
        feature_metadata_table = tables.get(
            "featureMetadata",
            f"{dataset_template.name.replace(' ', '_')}_feature_metadata",
        )
        predictions_table = tables.get(
            "predictions", f"{dataset_template.name.replace(' ', '_')}_predictions"
        )

        dataframe.to_sql(
            table_name=data_table, schema_name=database, if_exists="replace"
        )

        dataframe = dataframe.to_pandas()

        column_names = dataframe.columns
        column_dtypes = [str(dtype) for dtype in dataframe.dtypes]

        feature_columns_objects = [
            Variable()
            .set_name(col)
            .set_data_type(
                DataType.FLOAT if dtype.startswith("float") else DataType.INTEGER
            )
            .set_type(TypeEnum.FEATURE)
            for col, dtype in zip(column_names, column_dtypes)
            if col not in entity_columns + target_columns
        ]

        entity_columns_objects = [
            Variable()
            .set_name(col)
            .set_data_type(
                DataType.FLOAT
                if str(dataframe[col].dtype).startswith("float")
                else DataType.INTEGER
            )
            .set_type(TypeEnum.ENTITY)
            for col in entity_columns
        ]

        target_columns_objects = [
            Variable()
            .set_name(col)
            .set_data_type(
                DataType.FLOAT
                if str(dataframe[col].dtype).startswith("float")
                else DataType.INTEGER
            )
            .set_type(TypeEnum.TARGET)
            for col in target_columns
        ]

        self._create_tables(
            target_columns_objects,
            entity_columns,
            database,
            predictions_table,
            feature_metadata_table,
        )

        feature_metadata, metadata = self._build_metadata(
            entity_columns,
            target_columns,
            feature_columns_objects,
            entity_columns_objects,
            target_columns_objects,
            sql,
            database,
            data_table,
            entity_target_table,
            predictions_table,
            feature_metadata_table,
        )

        dataset_template.feature_metadata = feature_metadata
        dataset_template.metadata = metadata

        template_request = {
            "name": dataset_template.name,
            "description": dataset_template.description,
            "catalogType": dataset_template.catalog_type.value,
            "metadata": {
                "features": {
                    "sql": dataset_template.metadata.features.sql,
                    "entity": dataset_template.metadata.features.entity,
                    "variables": [
                        {
                            "name": col.name,
                            "dataType": col.data_type.value,
                            "type": col.type.value,
                        }
                        for col in dataset_template.metadata.features.variables
                    ],
                },
                "entityAndTargets": {
                    "entity": dataset_template.entity,
                    "sql": dataset_template.metadata.entity_and_targets.sql,
                    "variables": [
                        {
                            "name": col.name,
                            "dataType": col.data_type.value,
                            "type": col.type.value,
                        }
                        for col in dataset_template.metadata.entity_and_targets.variables
                    ],
                },
                "predictions": {
                    "database": dataset_template.metadata.predictions.database,
                    "entitySql": dataset_template.metadata.predictions.entity_sql,
                    "table": dataset_template.metadata.predictions.table,
                },
                "type": dataset_template.metadata.type.value,
            },
            "featureMetadata": {
                "database": dataset_template.feature_metadata.database,
                "table": dataset_template.feature_metadata.table,
            },
        }

        response = self.tmo_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=template_request,
        )

        dataset_template_id = uuid.UUID(response["id"])

        dataset_template.id = dataset_template_id

        logger.debug("Dataset template created successfully.")

        return dataset_template

    def render(self, id: str | uuid.UUID) -> dict:
        """
        returns a rendered dataset template

        Parameters:
           id (str): dataset_template id

        Returns:
            (dict): rendered dataset template
        """
        return self.tmo_client.get_request(
            path=f"{self.path}/{str(id)}/render",
            header_params=self._get_header_params(),
            query_params={},
        )

    def find_by_name_like(
        self, name: str, projection: str = None, return_dataframe: bool = False
    ) -> None | list[Any] | DataFrame:
        """
        returns datasets matching the name as a combined DataFrame or list

        Parameters:
            name (str): dataset name(string) to find
            projection (str): projection type
            return_dataframe (bool): if True, returns combined DataFrame; if False, returns list of Dataset objects

        Returns:
            (list): dataset template
        """
        query_vars = ["name", "projection"]
        query_vals = [name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        response = self.tmo_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        try:
            dataset_templates_data = response.get("_embedded", {}).get(
                "datasetTemplates", []
            )

            if not dataset_templates_data:
                logger.info(f"No dataset templates found for name: {name}")
                return None if return_dataframe else []

            dataset_templates = []
            for data in dataset_templates_data:
                dataset_template = self._parse_dataset_template_from_response(data)
                if dataset_template:
                    dataset_templates.append(dataset_template)

            if return_dataframe:
                return self._dataset_templates_to_dataframe(dataset_templates)
            else:
                return dataset_templates

        except Exception as e:
            logger.error(f"Error parsing dataset templates: {str(e)}")
            return None if return_dataframe else []

    def find_by_id(
        self, dataset_template_id: Union[str, uuid.UUID], return_dataframe: bool = False
    ) -> None | DataFrame | DatasetTemplate:
        """
        returns a dataset template by id

        Parameters:
        id (str): dataset_template id
        return_dataframe (bool): if True, returns DataFrame; if False, returns DatasetTemplate object

        Returns:
            (dict): dataset template
            (DatasetTemplate): dataset template object if return_dataframe is True
        """

        if not isinstance(dataset_template_id, uuid.UUID):
            try:
                dataset_template_id = uuid.UUID(dataset_template_id)
            except ValueError:
                logger.error(f"Invalid UUID format for id: {dataset_template_id}")
                return None

        query_vars = ["id"]
        query_vals = [str(dataset_template_id)]
        query_params = self.generate_params(query_vars, query_vals)

        response = self.tmo_client.get_request(
            path=f"{self.path}/search/findById",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        try:
            dataset_template = self._parse_dataset_template_from_response(response)

            if dataset_template is None:
                logger.error(
                    f"Dataset template with id {dataset_template_id} not found."
                )
                return None

            if return_dataframe:
                return dataset_template.to_dataframe()
            else:
                return dataset_template

        except Exception as e:
            logger.error(f"Error parsing dataset template: {str(e)}")
            return None

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
        return_dataframe: bool = False,
    ) -> list[Any] | DataFrame:
        """
        returns all dataset templates

        Parameters:
        projection (str): projection type
        page (int): page number
        size (int): number of records in a page
        sort (str): column name and sorting order
        return_dataframe (bool): if True, returns DataFrame; if False, returns list of DatasetTemplate objects

        Returns:
            (list): dataset templates
        """

        query_vars = ["projection", "page", "size", "sort"]
        query_vals = [projection, page, size, sort]
        built_query_params = self.generate_params(query_vars, query_vals)

        response = self.tmo_client.get_request(
            path=f"{self.path}",
            header_params=self._get_header_params(),
            query_params=built_query_params,
        )

        try:
            dataset_templates_data = response.get("_embedded", {}).get(
                "datasetTemplates", []
            )

            if not dataset_templates_data:
                logger.info("No dataset templates found.")
                return [] if not return_dataframe else pd.DataFrame()

            dataset_templates = []
            for data in dataset_templates_data:
                dataset_template = self._parse_dataset_template_from_response(data)
                if dataset_template:
                    dataset_templates.append(dataset_template)

            if return_dataframe:
                return self._dataset_templates_to_dataframe(dataset_templates)
            else:
                return dataset_templates

        except Exception as e:
            logger.error(f"Error parsing dataset templates: {str(e)}")
            return [] if not return_dataframe else pd.DataFrame()

    def _get_header_params(self) -> dict:
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    @staticmethod
    def _parse_dataset_template_from_response(response: dict) -> DatasetTemplate | None:
        try:
            dataset_template = DatasetTemplate(
                id=uuid.UUID(response.get("id")),
                name=response.get("name"),
                description=response.get("description"),
                project_id=uuid.UUID(response.get("projectId")),
                owner_id=response.get("ownerId"),
                catalog_type=CatalogType(response.get("catalogType")),
            )

            # Get metadata
            metadata = response.get("metadata")
            features_data = metadata.get("features")
            entity_targets_data = metadata.get("entityAndTargets")

            # Extract entity
            entity = features_data.get("entity")
            dataset_template.entity = entity

            # Extract targets
            if "variables" in entity_targets_data:
                target_variables = []
                for variable in entity_targets_data.get("variables"):
                    if variable.get("type") == TypeEnum.TARGET.value:
                        target_variables.append(variable.get("name"))
                dataset_template.target = target_variables

            # Set up the metadata structure
            dataset_template.metadata = Metadata()
            dataset_template.metadata.features = FeaturesEntityTargets()
            dataset_template.metadata.entity_and_targets = FeaturesEntityTargets()
            dataset_template.metadata.predictions = Predictions()
            dataset_template.metadata.type = CatalogBodyType(metadata.get("type"))

            entity_target_columns = DatasetTemplateApi._dict_list_to_variable_list(
                entity_targets_data.get("variables")
            )

            dataset_template.metadata.entity_and_targets.variables = (
                entity_target_columns
            )
            dataset_template.metadata.entity_and_targets.entity = entity
            dataset_template.metadata.entity_and_targets.sql = entity_targets_data.get(
                "sql"
            )

            feature_columns = DatasetTemplateApi._dict_list_to_variable_list(
                features_data.get("variables")
            )

            dataset_template.metadata.features.variables = feature_columns
            dataset_template.metadata.features.entity = entity
            dataset_template.metadata.features.sql = features_data.get("sql")

            predictions_data = metadata.get("predictions")
            dataset_template.metadata.predictions.database = predictions_data.get(
                "database"
            )
            dataset_template.metadata.predictions.entity_sql = predictions_data.get(
                "entitySql"
            )
            dataset_template.metadata.predictions.table = predictions_data.get("table")

            return dataset_template
        except Exception as e:
            logger.error(f"Error parsing dataset template from response: {str(e)}")
            return None

    @staticmethod
    def _dataset_templates_to_dataframe(
        dataset_templates: list[DatasetTemplate],
    ) -> DataFrame | None:
        """
        Helper method to combine multiple dataset templates into a single DataFrame

        Parameters:
            dataset_templates (list[DatasetTemplate]): a list of dataset templates to combine

        Returns:
            DataFrame: Combined DataFrame containing all dataset templates, or None if conversion fails
        """
        if not dataset_templates:
            return None

        try:
            from tmo.util.utils import to_dataframe

            combined_data = []
            for dataset_template in dataset_templates:
                combined_data.append(dataset_template.get_df_template())

            return to_dataframe(combined_data)
        except Exception as e:
            logger.error(f"Could not combine dataset templates to DataFrame: {str(e)}")
            return None

    @staticmethod
    def _dict_list_to_variable_list(variables: list[dict]) -> list[Variable]:
        return [
            Variable()
            .set_name(var["name"])
            .set_data_type(DataType(var["dataType"]))
            .set_type(TypeEnum(var["type"]))
            for var in variables
        ]

    @staticmethod
    def _create_tables(
        target_columns_objects: list[Variable],
        entity_columns: list[str],
        database: str,
        predictions_table: str,
        feature_metadata_table: str,
    ):
        # Create the predictions table schema
        predictions_table_schema = ", ".join([
            f"{col.name} {'FLOAT' if col.data_type == DataType.FLOAT else 'INTEGER'}"
            for col in target_columns_objects
        ])

        predictions_table_schema = (
            f"job_id VARCHAR(128), {entity_columns[0]} VARCHAR(128),"
            f" {predictions_table_schema}, json_report CLOB"
        )

        create_predictions_table_query = f"""CREATE TABLE "{database}"."{predictions_table}" ({predictions_table_schema});"""

        # Create the predictions table if it doesn't exist
        try:
            execute_sql(f'SELECT TOP 1 * FROM "{database}"."{predictions_table}";')
        except:  # noqa #NOSONAR
            try:
                execute_sql(create_predictions_table_query)
                warnings.warn(
                    f"Table {predictions_table} already exists. Using existing table.",
                    UserWarning,
                )
            except Exception as e:
                raise EntityCreationError(
                    f"Error creating table {predictions_table}: {e}"
                )

        # Create the feature metadata table if it doesn't exist
        try:
            execute_sql(f'SELECT TOP 1 * FROM "{database}"."{feature_metadata_table}";')
            warnings.warn(
                f"Table {feature_metadata_table} already exists. Using existing table.",
                UserWarning,
            )
        except:  # noqa #NOSONAR
            try:
                from ..stats.store import create_features_stats_table

                create_features_stats_table(feature_metadata_table)
            except Exception as e:
                raise EntityCreationError(
                    f"Error creating table {feature_metadata_table}: {e}"
                )

    @staticmethod
    def _build_metadata(
        entity_columns: list[str],
        target_columns: list[str],
        feature_columns_objects: list[Variable],
        entity_columns_objects: list[Variable],
        target_columns_objects: list[Variable],
        sql: Optional[dict],
        database: str,
        data_table: str,
        entity_target_table: str,
        predictions_table: str,
        feature_metadata_table: str,
    ) -> tuple[FeatureMetadata, Metadata]:
        # Build the 'features' metadata section
        features = FeaturesEntityTargets()
        features.entity = entity_columns[0]

        if sql and "features" in sql:
            logger.debug(f"Using custom SQL for features section: {sql['features']}")
            features.sql = sql["features"]
        else:
            features_table = data_table
            query = (
                f"SELECT {entity_columns[0]},"
                f" {', '.join([col.name for col in feature_columns_objects])} FROM"
                f" {features_table}"
            )
            logger.debug(f"Using data table for features SQL: {features_table}")
            features.sql = query
        features.variables = feature_columns_objects

        # Build the 'entityAndTargets' metadata section (entity + target columns)
        entity_and_targets = FeaturesEntityTargets(entity=entity_columns[0])

        if sql and "target_entity" in sql:
            logger.debug(
                f"Using custom SQL for entityAndTargets section: {sql['target_entity']}"
            )
            entity_and_targets.set_sql(sql["target_entity"])
        else:
            logger.debug(
                f"Using data table for entityAndTargets SQL: {entity_target_table}"
            )
            query = (
                f"SELECT {entity_columns[0]}, {', '.join(target_columns)} FROM"
                f" {entity_target_table}"
            )
            entity_and_targets.set_sql(query)
        entity_and_targets.set_variables(
            entity_columns_objects + target_columns_objects
        )

        # Create predictions object with custom SQL if available
        predictions = Predictions(database=database, table=predictions_table)
        if sql and "predictions" in sql:
            logger.debug(f"Using custom SQL for predictions: {sql['predictions']}")
            predictions.set_entity_sql(sql["predictions"])
        else:
            query = (
                f"SELECT {entity_columns[0]}, {', '.join(target_columns)} FROM"
                f" {data_table}"
            )
            logger.debug(
                "Custom SQL for predictions not provided. Using auto-generated SQL:"
                f" {query}"
            )
            predictions.set_entity_sql(query)

        feature_metadata = FeatureMetadata(database, feature_metadata_table)

        metadata = Metadata(
            type=CatalogBodyType.VANTAGE,
            predictions=predictions,
            entity_and_targets=entity_and_targets,
            features=features,
        )

        return feature_metadata, metadata



import os
import uuid
from typing import Generator, Any

import pandas
import pandas as pd
from teradataml import DataFrame, copy_to_sql


def to_dataframe(obj: dict | list) -> DataFrame | pandas.DataFrame:
    """
    Converts an object (list or dict) into a Teradata DataFrame.
    Falls back to returning a Pandas DataFrame if the Teradata conversion fails.

    Parameters:
        obj (dict, list): The object from the response to convert.

    Returns:
        teradataml.DataFrame | pandas.DataFrame: Teradata DataFrame if successful, otherwise Pandas DataFrame.
    """
    if obj is None:
        raise ValueError("Response object cannot be None")

    items = []
    if isinstance(obj, dict):
        if not any(isinstance(v, (dict, list)) for v in obj.values()):
            items = [obj]
    elif isinstance(obj, list):
        items = obj
    else:
        raise ValueError(f"Unsupported response_object type: {type(obj)}")

    if not items:
        raise ValueError("No items found in the response object")

    if not isinstance(items[0], dict):
        items = [{"value": item} for item in items]

    _normalized_keys, df_dict = _normalize_dict(items)

    pd_df = pd.DataFrame(df_dict)

    table_name = f"temp_modelops_table_{uuid.uuid4().hex[:8]}"

    try:
        copy_to_sql(pd_df, table_name=table_name, temporary=True)
        df = DataFrame.from_table(table_name)
        return df

    except Exception as e:
        print(f"[Fallback] Teradata DataFrame creation failed: {e}")
        print("[Fallback] Returning pandas DataFrame instead.")
        return pd_df


def get_files_in_path(path: str) -> Generator[str, Any, None]:
    for file in os.listdir(path):
        if os.path.isfile(os.path.join(path, file)):
            yield file


def _normalize_dict(items: list) -> tuple[dict, dict]:
    """
    Normalize keys from items and initialize/populate df_dict.

    Returns:
        normalized_keys: mapping original_key -> normalized_key
        df_dict: dict with normalized_key -> list of values per item
    """
    all_keys = set()
    for item in items:
        all_keys.update(item.keys())

    df_dict = {}
    normalized_keys = {}

    for key in all_keys:
        normalized_key = "".join(
            ["_" + c.lower() if c.isupper() else c for c in key]
        ).lstrip("_")
        normalized_keys[key] = normalized_key
        df_dict[normalized_key] = []

    for item in items:
        for original_key in all_keys:
            normalized_key = normalized_keys[original_key]
            df_dict[normalized_key].append(item.get(original_key, None))

    return normalized_keys, df_dict



import importlib
import json
import logging
import os
import sys

from tmo.cli.base_model import BaseModel
from tmo.cli.model_utils import cleanup_artifacts, ARTIFACTS_PATH
from tmo.context.model_context import ModelContext, DatasetInfo
from tmo.types.job import JobRunnerEngine, JobAutomationMode


class ScoreModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def batch_score_model_local(
        self, model_id: str = None, rendered_dataset: dict = None, base_path: str = None
    ):
        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = f"{base_path}model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(f"model directory {model_definitions_path} does not exist")

        model_artefacts_path = f".artefacts/{model_id}"

        if not os.path.exists(f"{model_artefacts_path}/output"):
            raise ValueError("You must run training before trying to run scoring.")

        try:
            cleanup_artifacts()

            os.makedirs(ARTIFACTS_PATH)

            os.symlink(
                f"../{model_artefacts_path}/output/",
                f"{ARTIFACTS_PATH}/input",
                target_is_directory=True,
            )
            os.symlink(
                "../{}/output/".format(model_artefacts_path),
                f"{ARTIFACTS_PATH}/output",
                target_is_directory=True,
            )

            def execute_score(
                engine,
                context,
                model_dir,
                model_definition,
                model_conf,
                rendered_dataset,
                cli_model_kwargs,
            ):
                if engine == JobRunnerEngine.PYTHON:
                    sys.path.append(model_dir)
                    scoring = importlib.import_module(
                        ".scoring", package="model_modules"
                    )

                    scoring.score(
                        context=context,
                        data_conf=rendered_dataset,
                        model_conf=model_conf,
                        **cli_model_kwargs,
                    )

                elif engine == JobRunnerEngine.SQL:
                    raise ValueError("SQL engine not supported")

                elif engine == JobRunnerEngine.R:
                    self._BaseModel__run_r_model(  # noqa
                        model_id, model_dir, rendered_dataset, "score.batch"
                    )

                else:
                    raise ValueError(
                        f"Unsupported language: {model_definition['language']}"
                    )

            self._execute_model(
                model_definitions_path=model_definitions_path,
                model_id=model_id,
                rendered_dataset=rendered_dataset,
                automation_mode=JobAutomationMode.DEPLOY,
                execute_callback=execute_score,
            )

            cleanup_artifacts()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            cleanup_artifacts()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                f"pip install -r {model_dir}/model_modules/requirements.txt"
            )
            raise
        except:
            cleanup_artifacts()
            self.logger.exception("Exception running model code")
            raise



from __future__ import absolute_import

import abc

from tmo.api_client import TmoClient


class BaseApi(object):
    path = ""
    json_type = "application/json"

    def __init__(self, tmo_client: TmoClient):
        self.tmo_client = tmo_client

    @abc.abstractmethod  # noqa
    def _get_header_params(self):
        pass

    def _get_standard_header_params(self, accept_types: list[str] = None):
        """
        Helper method to generate standard header parameters for API requests.

        Parameters:
            accept_types (list[str]): List of acceptable response types.
                                     If None, defaults to json_type only.

        Returns:
            (dict): generated header parameters
        """
        if accept_types is None:
            accept_value = self.json_type
        else:
            accept_value = self.tmo_client.select_header_accept(accept_types)

        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.tmo_client.project_id,
            self.tmo_client.project_id,
            self.json_type,
            accept_value,
        ]

        return self.generate_params(header_vars, header_vals)

    def _approve_entity(self, entity_id: str, comments: str):
        """
        Generic method to approve an entity (trained model, feature engineering task, etc.)

        Parameters:
            entity_id (str): entity id(uuid)
            comments (str): approval comments

        Returns:
            (dict): response
        """
        approve_request = {"comments": comments}

        return self.tmo_client.post_request(
            path=f"{self.path}/{entity_id}/approve",
            header_params=self._get_header_params(),
            query_params={},
            body=approve_request,
        )

    def _reject_entity(self, entity_id: str, comments: str):
        """
        Generic method to reject an entity (trained model, feature engineering task, etc.)

        Parameters:
            entity_id (str): entity id(uuid)
            comments (str): rejection comments

        Returns:
            (dict): response
        """
        reject_request = {"comments": comments}

        return self.tmo_client.post_request(
            path=f"{self.path}/{entity_id}/reject",
            header_params=self._get_header_params(),
            query_params={},
            body=reject_request,
        )

    def _deploy_entity(self, entity_id: str, deploy_request: dict):
        """
        Generic method to deploy an entity (trained model, feature engineering task, etc.)

        Parameters:
            entity_id (str): entity id(uuid)
            deploy_request (dict): deployment request

        Returns:
            (dict): response
        """
        self.required_params(["engineType"], deploy_request)

        return self.tmo_client.post_request(
            path=f"{self.path}/{entity_id}/deploy",
            header_params=self._get_header_params(),
            query_params={},
            body=deploy_request,
        )

    def _retire_entity(self, entity_id: str, retire_request: dict):
        """
        Generic method to retire an entity (trained model, feature engineering task, etc.)

        Parameters:
            entity_id (str): entity id(uuid)
            retire_request (dict): retire request

        Returns:
            (dict): response
        """
        self.required_params(["deploymentId"], retire_request)

        return self.tmo_client.post_request(
            path=f"{self.path}/{entity_id}/retire",
            header_params=self._get_header_params(),
            query_params={},
            body=retire_request,
        )

    @staticmethod
    def generate_params(params: list[str], values: list[str]):
        """
        returns list of parameters and values as dictionary

        Parameters:
           params (list[str]): list of parameter names
           values (list[str]): list of parameter values

        Returns:
            (dict): generated parameters
        """

        # bools in python start with upper case when converted to strs. APIs expect lowercase
        api_values = [str(v).lower() if type(v) is bool else v for v in values]

        return dict(zip(params, api_values))

    @staticmethod
    def required_params(param_names: list[str], dict_obj: dict[str, str]):
        """
        checks required parameters, raises exception if the required parameter is missing in the dictionary

        Parameters:
           param_names (list[str]): list of required parameter names
           dict_obj (Dict[str, str]): dictionary to check for required parameters
        """
        for param in param_names:
            if param not in dict_obj:
                raise ValueError(f"Missing required value {str(param)}")

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all entities

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ["projection", "page", "size", "sort"]
        query_vals = [projection, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(self.path, header_params, query_params)

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all entities by archived

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order e.g. name?asc / name?desc
           archived (bool): whether to return archived or unarchived entities

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ["projection", "page", "size", "sort", "archived"]
        query_vals = [projection, page, size, sort, archived]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            f"{self.path}/search/findByArchived", header_params, query_params
        )

    def find_by_id(self, id: str, projection: str = None):
        """
        returns the entity

        Parameters:
           id (str): entity id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            f"{self.path}/{id}", header_params, query_params
        )

    def archive(self, id: str):
        """
        archives the entity
        Parameters:
           id (str): entity id(uuid) to archive
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.tmo_client.post_request(
            f"/api/archives/{self.type}/{id}", header_params, {}, {}
        )

    def unarchive(self, id: str):
        """
        unarchives the entity
        Parameters:
           id (str): entity id(uuid) to unarchive
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.tmo_client.delete_request(
            f"/api/archives/{self.type}/{id}", header_params, {}, {}
        )



from __future__ import absolute_import

from tmo.api.iterator_base_api import IteratorBaseApi


class JobApi(IteratorBaseApi):
    path = "/api/jobs"
    type = "JOB"

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        raise NotImplementedError("Archiving not supported for Jobs")

    def _get_header_params(self):
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    def find_job_events(self, job_id: str, projection: str = None):
        """
        returns events of a job

        Parameters:
           job_id (str): job id(uuid) to find events
           projection (str): projection type

        Returns:
            (dict): job events
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/{job_id}/events",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_job_event_id(
        self, job_id: str, job_event_id: str, projection: str = None
    ):
        """
        returns job event

        Parameters:
           job_id (str): job id(uuid)
           job_event_id (str): job event id(uuid)
           projection (str): projection type

        Returns:
            (dict): job event
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/{job_id}/events/{job_event_id}",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_deployment_id(self, deployment_id: str, projection: str = None):
        """
        returns list of jobs for a given deployment

        Parameters:
           deployment_id (str): deployment id(string) to find
           projection (str): projection type

        Returns:
            (dict): jobs
        """
        query_vars = ["deploymentId", "projection"]
        query_vals = [deployment_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByDeploymentId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def wait(self, job_id: str, timeout_sec: int = 60):
        import time

        start_time_sec = time.time()

        while True:
            job = self.find_by_id(id=job_id, projection="expandJob")
            status = (
                job.get("status", "UNKNOWN") if isinstance(job, dict) else "UNKNOWN"
            )

            if status == "COMPLETED":
                return
            elif status in ["ERROR", "CANCELLED"]:
                raise SystemError(f"Job failed with status: {status}")

            elapsed = time.time() - start_time_sec
            if elapsed > timeout_sec:
                raise TimeoutError(
                    f"Timeout waiting for job to complete. Current status: {status}"
                )

            time.sleep(5)



import builtins
import json
import os
from pathlib import Path

import pytest
import yaml

import tmo.tmo_cli as tmo_cli
from types import SimpleNamespace


def test_bash_escape_handles_backslashes_and_none():
    assert tmo_cli.bash_escape(r"C:\path\to\file") == r"C:\\path\\to\\file"
    assert tmo_cli.bash_escape("") == ""
    assert tmo_cli.bash_escape(None) is None


def test_print_underscored_outputs_message_and_underline(capsys):
    tmo_cli.print_underscored("Hello")
    out, _ = capsys.readouterr()
    lines = out.splitlines()
    assert lines[0] == "Hello"
    assert lines[1] == "-" * len("Hello")


def test_yes_or_no_returns_true_and_false(monkeypatch):
    # simulate answering 'y'
    monkeypatch.setattr(builtins, "input", lambda prompt="": "y")
    assert tmo_cli.yes_or_no("question") is True

    # simulate answering 'n'
    monkeypatch.setattr(builtins, "input", lambda prompt="": "n")
    assert tmo_cli.yes_or_no("question") is False


def test_set_cwd_nonexistent_path_exits(monkeypatch):
    # Force os.path.exists to return False inside module
    monkeypatch.setattr(tmo_cli.os.path, "exists", lambda path: False)
    with pytest.raises(SystemExit) as exc:
        tmo_cli.set_cwd("/nonexistent/path")
    assert exc.value.code == 1


def test_handle_generic_error_debug_true_logs_exception(caplog):
    caplog.clear()
    # debug True should call logging.exception and not exit
    tmo_cli.handle_generic_error(Exception("boom"), debug=True)
    assert any(
        "An error occurred, printing stack trace output." in r.getMessage()
        or "An error occurred" in r.getMessage()
        for r in caplog.records
    )


def test_handle_generic_error_debug_false_exits_and_logs(caplog):
    caplog.clear()
    with pytest.raises(SystemExit) as exc:
        tmo_cli.handle_generic_error(Exception("boom"), debug=False)
    assert exc.value.code == 1
    assert any("An error occurred" in r.getMessage() for r in caplog.records)


def test_link_repo_calls_write_repo_config(monkeypatch, capsys):
    # Prepare a fake project returned by the selection function
    monkeypatch.setattr(
        tmo_cli,
        "list_and_select_projects",
        lambda _repo_manager, tmo_client, a, b: {"id": "proj-123"},
    )

    class DummyRepoManager:
        def __init__(self):
            self.written = None

        def write_repo_config(self, config, path=None):
            # mirror behavior: accept optional path param
            self.written = (config, path)

    repo_manager = DummyRepoManager()
    # call link_repo and capture stdout
    tmo_cli.link_repo(repo_manager, None)
    out, _ = capsys.readouterr()
    assert repo_manager.written == (
        {"project_id": "proj-123"},
    ) or repo_manager.written == ({"project_id": "proj-123"}, None)
    assert "Repo linked to Project." in out


def test_input_string_required_retry(monkeypatch, capsys):
    calls = {"count": 0}

    def fake_input(prompt=""):
        calls["count"] += 1
        if calls["count"] == 1:
            return ""
        return "value"

    monkeypatch.setattr("builtins.input", fake_input)
    val = tmo_cli.input_string("name", required=True)
    assert val == "value"


def test_input_string_password_getpass(monkeypatch):
    monkeypatch.setattr("getpass.getpass", lambda prompt=None: "secret")
    val = tmo_cli.input_string(
        "pwd", required=True, password=True, is_called_from_test=False
    )
    assert val == "secret"


def test_input_select_default_and_invalid(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda prompt="": "")
    values = ["a", "b"]
    res = tmo_cli.input_select("item", values, default="a")
    assert res == "a"
    # test invalid selection then valid
    seq = iter(["", "2", "1"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(seq))
    res2 = tmo_cli.input_select("item", values)
    assert res2 == "b"


def test_list_connections_no_file(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    args = SimpleNamespace(cwd=None)
    with pytest.raises(SystemExit):
        tmo_cli.list_connections(args)


def test_add_connections_writes_file(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    args = SimpleNamespace(
        cwd=None,
        name="conn1",
        username="u",
        password="p",
        host="h",
        database="db",
        val_db="VAL",
        byom_db="MLDB",
        logmech="TDNEGO",
        parent_parser=None,
    )
    monkeypatch.setattr(tmo_cli.crypto, "td_encrypt_password", lambda **kwargs: "ENC")
    tmo_cli.add_connections(args)
    f = Path(tmo_cli.config_dir) / "connections.yaml"
    assert f.exists()
    assert "connections" in yaml.safe_load(open(f))


def test_remove_connections_no_file(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    args = SimpleNamespace(cwd=None, connection=None)
    with pytest.raises(SystemExit) as exc:
        tmo_cli.remove_connections(args)
    assert exc.value.code == 0


def test_remove_connections_not_exists(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    con = {
        "connections": [{
            "id": "abc",
            "name": "n",
            "username": "u",
            "password": "p",
            "host": "h",
            "logmech": "TDNEGO",
        }]
    }
    yaml.safe_dump(con, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"))
    args = SimpleNamespace(cwd=None, connection="notfound")
    with pytest.raises(SystemExit) as exc:
        tmo_cli.remove_connections(args)
    assert exc.value.code == 1


def test_remove_connections_success(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    conn_id = "c1"
    con = {
        "connections": [{
            "id": conn_id,
            "name": "n",
            "username": "u",
            "password": "p",
            "host": "h",
            "logmech": "TDNEGO",
        }]
    }
    yaml.safe_dump(con, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"))
    Path(tmo_cli.config_dir, f"{conn_id}.key").write_text("k")
    Path(tmo_cli.config_dir, f"{conn_id}.pass").write_text("p")
    args = SimpleNamespace(cwd=None, connection=None)
    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "n")
    tmo_cli.remove_connections(args)
    data = yaml.safe_load(open(Path(tmo_cli.config_dir) / "connections.yaml"))
    assert data.get("connections", []) == []
    assert not Path(tmo_cli.config_dir, f"{conn_id}.key").exists()


def test_export_connection_prints(tmp_path, monkeypatch, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connection = {
        "id": "c1",
        "name": "n",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TDNEGO",
        "database": "db",
        "val_db": "VAL",
        "byom_db": "BYOM",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    args = SimpleNamespace(cwd=None, connection=None)
    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "n")
    tmo_cli.export_connection(args)


def test_activate_connection_sets_env_and_returns(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connection = {
        "id": "c1",
        "name": "n",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TDNEGO",
        "database": "db",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    args = SimpleNamespace(cwd=None, connection="c1")
    ret = tmo_cli.activate_connection(args)
    assert ret == "c1"
    assert os.environ.get("VMO_CONN_USERNAME") == "u"


def test_activate_connection_auto_select(monkeypatch, tmp_path, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connection = {
        "id": "c1",
        "name": "n",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TDNEGO",
        "database": "db",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    args = SimpleNamespace(cwd=None, connection=None)
    ret = tmo_cli.activate_connection(args)
    assert ret == "c1"


def test_test_connection_success(monkeypatch, tmp_path, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connection = {
        "id": "c1",
        "name": "n",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TDNEGO",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    class R:
        def fetchall(self):
            return [("ver1",)]

    monkeypatch.setattr(tmo, "execute_sql", lambda q: R())
    args = SimpleNamespace(cwd=None, connection="c1")
    tmo_cli.test_connection(args)


def test_test_connection_failure(monkeypatch, tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connection = {
        "id": "c1",
        "name": "n",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TDNEGO",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        tmo, "execute_sql", lambda q: (_ for _ in ()).throw(Exception("fail"))
    )
    args = SimpleNamespace(cwd=None, connection="c1")
    with pytest.raises(SystemExit):
        tmo_cli.test_connection(args)


def test_create_byom_table_no_execute(tmp_path):
    args = SimpleNamespace(cwd=None, name="tbl", execute_ddl=False)
    tmo_cli.create_byom_table(args)


def test_create_byom_table_execute_error(monkeypatch, tmp_path):
    args = SimpleNamespace(cwd=None, name="tbl", execute_ddl=True)
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        tmo, "execute_sql", lambda q: (_ for _ in ()).throw(Exception("boom"))
    )
    with pytest.raises(tmo_cli.EntityCreationError):
        tmo_cli.create_byom_table(args)


def test_set_cwd_success(tmp_path):
    d = tmp_path / "repo"
    d.mkdir()
    tmo_cli.set_cwd(str(d))
    assert tmo_cli.base_path == str(d.resolve())
    assert tmo_cli.model_catalog.endswith(tmo_cli.MODEL_CATALOG_PATH + "/")


def test_print_help_shows_version(capsys):
    ns = SimpleNamespace(version=True)
    tmo_cli.print_help(ns, parent_parser=None)
    out, _ = capsys.readouterr()
    from tmo import __version__

    assert __version__ in out


def test_list_and_select_projects_as_list(monkeypatch, capsys):
    monkeypatch.setattr("tmo.ProjectApi", lambda tmo_client, show_archived=False: [])
    ret = tmo_cli.list_and_select_projects(None, None, as_list=True, check_config=False)
    assert ret is None


def test_get_current_project_found(monkeypatch):
    class RM:
        def read_repo_config(self):
            return {"project_id": "p1"}

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_by_id(self, pid):
            return {"id": pid, "name": "proj"}

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)
    res = tmo_cli.get_current_project(RM(), None, check_repo_conf=False)
    assert res["id"] == "p1"


def test_validate_model_and_fe_tasks_cwd_valid(tmp_path, monkeypatch):
    d_model = tmp_path / "models"
    d_model.mkdir(parents=True)
    tmo_cli.model_catalog = str(d_model) + "/"
    assert tmo_cli.validate_model_catalog_cwd_valid() is True
    d_model.rmdir()
    assert tmo_cli.validate_model_catalog_cwd_valid() is False

    d_tasks = tmp_path / "fe_tasks"
    d_tasks.mkdir(parents=True)
    tmo_cli.fe_tasks_catalog = str(d_tasks) + "/"
    assert tmo_cli.validate_fe_tasks_cwd_valid() is True
    d_tasks.rmdir()
    assert tmo_cli.validate_fe_tasks_cwd_valid() is False


def test_init_model_directory_calls_link_when_no_repo_config(monkeypatch):
    class RM:
        def __init__(self):
            self.inited = False

        def init_model_directory(self):
            self.inited = True

        def repo_config_exists(self):
            return False

    repo_manager = RM()
    called = {"link": False}
    monkeypatch.setattr(
        tmo_cli,
        "link_repo",
        lambda repo_manager, tmo_client: called.update({"link": True}),
    )
    args = type("A", (), {"cwd": None})()
    tmo_cli.init_model_directory(args, repo_manager, None)
    assert repo_manager.inited is True
    assert called["link"] is True


def test_add_model_templates_empty_exits(monkeypatch, tmp_path):
    args = type("A", (), {"cwd": None, "template_url": "u", "branch": "b"})()

    class RM:
        def clone_repository(self, url, path, branch):
            return None

        def get_templates(self, entity_type=None, source_path=None):
            return {}

        def repo_config_exists(self):
            return True

    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)
    with pytest.raises(SystemExit):
        tmo_cli.add_model(args, RM())


def test_run_model_no_models_exits(monkeypatch):
    args = type(
        "A",
        (),
        {
            "cwd": None,
            "model_id": None,
            "mode": None,
            "local_dataset": None,
            "local_dataset_template": None,
            "dataset_id": None,
            "dataset_template_id": None,
            "connection": None,
        },
    )()
    monkeypatch.setattr(
        tmo_cli,
        "get_current_project",
        lambda repo_manager, tmo_client, check: {"id": "p1"},
    )
    import tmo

    class Client:
        def set_project_id(self, pid):
            pass

    monkeypatch.setattr(tmo.TrainModel, "get_model_ids", lambda catalog, arg: {})
    with pytest.raises(SystemExit):
        tmo_cli.run_model(args, None, Client())


def test_list_resources_invalid_selection_calls_help_and_exits(monkeypatch):
    args = type(
        "A",
        (),
        {
            "cwd": None,
            "projects": False,
            "models": False,
            "local_models": False,
            "templates": False,
            "datasets": False,
            "connections": False,
        },
    )()
    parent = type("P", (), {"print_help": lambda self: None})()
    with pytest.raises(SystemExit):
        tmo_cli.list_resources(args, None, None, parent_parser=parent)


def test_input_select_empty_values_returns_none():
    result = tmo_cli.input_select("test", [])
    assert result is None


def test_input_select_numeric_validation(monkeypatch):
    seq = iter(["abc", "0"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(seq))
    values = ["option1", "option2"]
    result = tmo_cli.input_select("item", values)
    assert result == "option1"


def test_yes_or_no_invalid_input_retry(monkeypatch):
    seq = iter(["maybe", "x", "y"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(seq))
    result = tmo_cli.yes_or_no("question")
    assert result is True


def test_bash_escape_with_multiple_backslashes():
    assert tmo_cli.bash_escape(r"C:\path\to\file\dir") == r"C:\\path\\to\\file\\dir"


def test_input_string_empty_not_required():
    import builtins
    import tmo.tmo_cli

    original_input = builtins.input
    builtins.input = lambda prompt="": ""
    result = tmo_cli.input_string("test", required=False)
    builtins.input = original_input
    assert result == ""


def test_input_string_tooltip_displays(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda prompt="": "value")
    tmo_cli.input_string("test", tooltip="This is a tooltip")
    out, _ = capsys.readouterr()
    assert "This is a tooltip" in out


def test_set_cwd_updates_global_paths(tmp_path):
    test_dir = tmp_path / "test_repo"
    test_dir.mkdir()
    tmo_cli.set_cwd(str(test_dir))
    assert tmo_cli.base_path == str(test_dir.resolve())
    assert tmo_cli.model_catalog == str(test_dir.resolve()) + "/model_definitions/"
    assert (
        tmo_cli.fe_tasks_catalog
        == str(test_dir.resolve()) + "/feature_engineering_tasks/"
    )


def test_list_connections_with_data(tmp_path, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connections = {
        "connections": [
            {
                "id": "c1",
                "name": "conn1",
                "username": "user1",
                "host": "host1",
                "database": "db1",
            },
            {
                "id": "c2",
                "name": "conn2",
                "username": "user2",
                "host": "host2",
            },
        ]
    }
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )
    args = SimpleNamespace(cwd=None)
    tmo_cli.list_connections(args)
    out, _ = capsys.readouterr()
    assert "conn1" in out
    assert "conn2" in out


def test_add_connections_without_args_prompts_user(monkeypatch, tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)

    inputs = iter(["testconn", "testuser", "testhost", "BYOM", "VAL", "testdb", "TD2"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(inputs))
    monkeypatch.setattr("getpass.getpass", lambda prompt="": "testpass")
    monkeypatch.setattr(
        tmo_cli.crypto, "td_encrypt_password", lambda **kwargs: "ENCRYPTED"
    )

    args = SimpleNamespace(
        cwd=None,
        name=None,
        username=None,
        password=None,
        host=None,
        database=None,
        val_db=None,
        byom_db=None,
        logmech=None,
        parent_parser=None,
    )

    tmo_cli.add_connections(args)

    f = Path(tmo_cli.config_dir) / "connections.yaml"
    assert f.exists()
    data = yaml.safe_load(open(f))
    assert len(data["connections"]) == 1
    assert data["connections"][0]["name"] == "testconn"


def test_add_connections_partial_args_exits(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)

    args = SimpleNamespace(
        cwd=None,
        name="test",
        username=None,
        password=None,
        host=None,
        database=None,
        val_db=None,
        byom_db=None,
        logmech=None,
        parent_parser=type("P", (), {"print_help": lambda self: None})(),
    )

    with pytest.raises(SystemExit):
        tmo_cli.add_connections(args)


def test_add_connections_existing_file_appends(monkeypatch, tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    existing = {"connections": [{"id": "old", "name": "old_conn"}]}
    yaml.safe_dump(existing, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"))

    monkeypatch.setattr(tmo_cli.crypto, "td_encrypt_password", lambda **kwargs: "ENC")

    args = SimpleNamespace(
        cwd=None,
        name="new_conn",
        username="u",
        password="p",
        host="h",
        database="db",
        val_db="VAL",
        byom_db="BYOM",
        logmech="TD2",
        parent_parser=None,
    )

    tmo_cli.add_connections(args)

    data = yaml.safe_load(open(Path(tmo_cli.config_dir) / "connections.yaml"))
    assert len(data["connections"]) == 2


def test_remove_connections_with_args(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    conn_id = "c1"
    connections = {
        "connections": [{
            "id": conn_id,
            "name": "conn1",
            "username": "u",
            "password": "p",
            "host": "h",
            "logmech": "TD2",
        }]
    }
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    args = SimpleNamespace(cwd=None, connection=conn_id)
    tmo_cli.remove_connections(args)

    data = yaml.safe_load(open(Path(tmo_cli.config_dir) / "connections.yaml"))
    assert len(data.get("connections", [])) == 0


def test_remove_connections_with_select(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    conn_id = "c1"
    connections = {
        "connections": [{
            "id": conn_id,
            "name": "conn1",
            "username": "u",
            "password": "p",
            "host": "h",
            "logmech": "TD2",
        }]
    }
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    args = SimpleNamespace(cwd=None, connection=None)
    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "conn1")

    tmo_cli.remove_connections(args)

    data = yaml.safe_load(open(Path(tmo_cli.config_dir) / "connections.yaml"))
    assert len(data.get("connections", [])) == 0


def test_remove_connections_empty_list(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connections = {"connections": []}
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    args = SimpleNamespace(cwd=None, connection=None)
    tmo_cli.remove_connections(args)


def test_export_connection_no_file(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)

    args = SimpleNamespace(cwd=None, connection=None)
    with pytest.raises(SystemExit):
        tmo_cli.export_connection(args)


def test_export_connection_not_found(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connections = {"connections": [{"id": "c1", "name": "conn1"}]}
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    args = SimpleNamespace(cwd=None, connection="notfound")
    with pytest.raises(SystemExit):
        tmo_cli.export_connection(args)


def test_export_connection_with_select(tmp_path, monkeypatch, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connection = {
        "id": "c1",
        "name": "conn1",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TD2",
        "database": "db",
        "val_db": "VAL",
        "byom_db": "BYOM",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )

    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "conn1")
    args = SimpleNamespace(cwd=None, connection=None)
    tmo_cli.export_connection(args)
    out, _ = capsys.readouterr()
    assert "export VMO_CONN_USERNAME" in out
    assert "u" in out


def test_export_connection_with_args(tmp_path, monkeypatch, capsys):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connection = {
        "id": "c1",
        "name": "conn1",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TD2",
        "database": "db",
        "val_db": "VAL",
        "byom_db": "BYOM",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )

    args = SimpleNamespace(cwd=None, connection="c1")
    tmo_cli.export_connection(args)
    out, _ = capsys.readouterr()
    assert "export VMO_CONN_USERNAME" in out
    assert "u" in out


def test_activate_connection_not_found(tmp_path, monkeypatch):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connections = {
        "connections": [
            {
                "id": "c1",
                "name": "conn1",
                "username": "u",
                "password": "p",
                "host": "h",
                "logmech": "TD2",
                "database": "db",
            },
            {
                "id": "c2",
                "name": "conn2",
                "username": "u2",
                "password": "p2",
                "host": "h2",
                "logmech": "TD2",
                "database": "db2",
            },
        ]
    }
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "notfound")

    args = SimpleNamespace(cwd=None, connection=None)
    with pytest.raises(SystemExit):
        tmo_cli.activate_connection(args)


def test_activate_connection_no_file(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)

    args = SimpleNamespace(cwd=None, connection=None)
    with pytest.raises(SystemExit):
        tmo_cli.activate_connection(args)


def test_activate_connection_with_kwargs(tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connection = {
        "id": "c1",
        "name": "conn1",
        "username": "u",
        "password": "p",
        "host": "h",
        "logmech": "TD2",
        "database": "db",
    }
    yaml.safe_dump(
        {"connections": [connection]},
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )

    args = SimpleNamespace(cwd=None, connection=None)
    ret = tmo_cli.activate_connection(args, connection="c1")
    assert ret == "c1"


def test_activate_connection_multiple_selection(monkeypatch, tmp_path):
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)

    connections = {
        "connections": [
            {
                "id": "c1",
                "name": "conn1",
                "username": "u1",
                "password": "p1",
                "host": "h1",
                "logmech": "TD2",
                "database": "db1",
            },
            {
                "id": "c2",
                "name": "conn2",
                "username": "u2",
                "password": "p2",
                "host": "h2",
                "logmech": "TD2",
                "database": "db2",
            },
        ]
    }
    yaml.safe_dump(
        connections, open(Path(tmo_cli.config_dir) / "connections.yaml", "w+")
    )

    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "conn2")
    args = SimpleNamespace(cwd=None, connection=None)
    ret = tmo_cli.activate_connection(args)
    assert ret == "c2"
    assert os.environ.get("VMO_CONN_USERNAME") == "u2"


def test_create_byom_table_execute_success(monkeypatch, capsys):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(tmo, "execute_sql", lambda q: None)

    args = SimpleNamespace(cwd=None, name="test_table", execute_ddl=True)
    tmo_cli.create_byom_table(args)
    out, _ = capsys.readouterr()
    assert "created successfully" in out


def test_compute_stats_categorical(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import stats, store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    class MockDF:
        pass

    monkeypatch.setattr("teradataml.DataFrame.from_query", lambda q: MockDF())

    monkeypatch.setattr(
        stats,
        "compute_categorical_stats",
        lambda df, cols, temp_db=None: {"col1": {"categories": ["a", "b"]}},
    )

    monkeypatch.setattr(store, "save_feature_stats", lambda **kwargs: None)

    args = SimpleNamespace(
        cwd=None,
        source_table="test.table",
        metadata_table="test.metadata",
        feature_type="categorical",
        columns="col1, col2",
        temp_view_database=None,
    )

    tmo_cli.compute_stats(args)


def test_compute_stats_continuous(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import stats, store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    class MockDF:
        pass

    monkeypatch.setattr("teradataml.DataFrame.from_query", lambda q: MockDF())

    monkeypatch.setattr(
        stats,
        "compute_continuous_stats",
        lambda df, cols, temp_db=None: {"col1": {"edges": [1, 2, 3]}},
    )

    monkeypatch.setattr(store, "save_feature_stats", lambda **kwargs: None)

    args = SimpleNamespace(
        cwd=None,
        source_table="test.table",
        metadata_table="test.metadata",
        feature_type="continuous",
        columns="col1, col2",
        temp_view_database=None,
    )

    tmo_cli.compute_stats(args)


def test_compute_stats_error(monkeypatch):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        "teradataml.DataFrame.from_query",
        lambda q: (_ for _ in ()).throw(Exception("error")),
    )

    args = SimpleNamespace(
        cwd=None,
        source_table="test.table",
        metadata_table="test.metadata",
        feature_type="continuous",
        columns="col1",
        temp_view_database=None,
    )

    with pytest.raises(RuntimeError):
        tmo_cli.compute_stats(args)


def test_compute_stats_with_temp_view_database(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import stats, store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    class MockDF:
        pass

    monkeypatch.setattr("teradataml.DataFrame.from_query", lambda q: MockDF())

    categorical_calls = []
    continuous_calls = []

    def mock_categorical_stats(df, cols, temp_db=None):
        categorical_calls.append({"df": df, "cols": cols, "temp_db": temp_db})
        return {"col1": {"categories": ["a", "b"]}}

    def mock_continuous_stats(df, cols, temp_db=None):
        continuous_calls.append({"df": df, "cols": cols, "temp_db": temp_db})
        return {"col1": {"edges": [1, 2, 3]}}

    monkeypatch.setattr(stats, "compute_categorical_stats", mock_categorical_stats)
    monkeypatch.setattr(stats, "compute_continuous_stats", mock_continuous_stats)
    monkeypatch.setattr(store, "save_feature_stats", lambda **kwargs: None)

    args = SimpleNamespace(
        cwd=None,
        source_table="test.table",
        metadata_table="test.metadata",
        feature_type="categorical",
        columns="col1, col2",
        temp_view_database="temp_view",
    )

    tmo_cli.compute_stats(args)

    assert len(categorical_calls) == 1
    assert categorical_calls[0]["cols"] == ["col1", "col2"]

    categorical_calls.clear()
    args.feature_type = "continuous"

    tmo_cli.compute_stats(args)

    assert len(continuous_calls) == 1
    assert continuous_calls[0]["cols"] == ["col1", "col2"]


def test_list_stats_success(monkeypatch, capsys):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        store,
        "get_feature_stats_summary",
        lambda table: {"col1": "continuous", "col2": "categorical"},
    )

    args = SimpleNamespace(cwd=None, metadata_table="test.metadata")
    tmo_cli.list_stats(args)
    out, _ = capsys.readouterr()
    assert "col1" in out
    assert "col2" in out


def test_list_stats_empty(monkeypatch):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(store, "get_feature_stats_summary", lambda table: {})

    args = SimpleNamespace(cwd=None, metadata_table="test.metadata")
    with pytest.raises(RuntimeError):
        tmo_cli.list_stats(args)


def test_list_stats_error(monkeypatch):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        store,
        "get_feature_stats_summary",
        lambda table: (_ for _ in ()).throw(Exception("error")),
    )

    args = SimpleNamespace(cwd=None, metadata_table="test.metadata")
    with pytest.raises(RuntimeError):
        tmo_cli.list_stats(args)


def test_create_stats_table_no_execute(capsys):
    args = SimpleNamespace(cwd=None, metadata_table="test.metadata", execute_ddl=False)
    tmo_cli.create_stats_table(args)
    out, _ = capsys.readouterr()
    assert "Execution not requested" in out


def test_create_stats_table_execute_success(monkeypatch, capsys):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(tmo, "execute_sql", lambda q: None)

    args = SimpleNamespace(cwd=None, metadata_table="test.metadata", execute_ddl=True)
    tmo_cli.create_stats_table(args)
    out, _ = capsys.readouterr()
    assert "created successfully" in out


def test_create_stats_table_error(monkeypatch):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        tmo, "execute_sql", lambda q: (_ for _ in ()).throw(Exception("error"))
    )

    args = SimpleNamespace(cwd=None, metadata_table="test.metadata", execute_ddl=True)
    with pytest.raises(RuntimeError):
        tmo_cli.create_stats_table(args)


def test_import_stats_show_example(capsys):
    args = SimpleNamespace(
        cwd=None, show_example=True, statistics_file=None, metadata_table=None
    )
    with pytest.raises(SystemExit):
        tmo_cli.import_stats(args)
    out, _ = capsys.readouterr()
    assert "age" in out
    assert "continuous" in out


def test_import_stats_success(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo
    from tmo.stats import store

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(store, "save_feature_stats", lambda *args, **kwargs: None)

    stats_file = tmp_path / "stats.json"
    stats_data = {
        "features": {
            "age": {"type": "continuous", "edges": [1, 2, 3]},
            "gender": {"type": "categorical", "categories": ["M", "F"]},
        }
    }
    stats_file.write_text(json.dumps(stats_data))

    args = SimpleNamespace(
        cwd=None,
        show_example=False,
        statistics_file=str(stats_file),
        metadata_table="test.metadata",
    )

    tmo_cli.import_stats(args)


def test_import_stats_error(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    import tmo

    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    stats_file = tmp_path / "stats.json"
    stats_file.write_text("invalid json")

    args = SimpleNamespace(
        cwd=None,
        show_example=False,
        statistics_file=str(stats_file),
        metadata_table="test.metadata",
    )

    with pytest.raises(RuntimeError):
        tmo_cli.import_stats(args)


def test_doctor_success(monkeypatch):
    monkeypatch.setattr(tmo_cli, "test_connection", lambda args: None)
    import tmo

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def __iter__(self):
            return iter([{"id": "p1", "name": "proj1"}])

    monkeypatch.setattr("tmo.ProjectApi", PApi)

    args = SimpleNamespace(cwd=None, connection=None)
    tmo_cli.doctor(args, None, None)


def test_doctor_no_projects(monkeypatch):
    import tmo
    from tmo.types.exceptions import EntityNotFoundError

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def __iter__(self):
            return iter([])

    monkeypatch.setattr("tmo.ProjectApi", PApi)

    args = SimpleNamespace(cwd=None, connection=None)
    with pytest.raises(EntityNotFoundError):
        tmo_cli.doctor(args, None, None)


def test_doctor_connection_error(monkeypatch):
    from tmo.types.exceptions import ConfigurationError

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            raise ConfigurationError("config error")

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "test_connection", lambda args: None)

    args = SimpleNamespace(cwd=None, connection=None)
    tmo_cli.doctor(args, None, None)


def test_print_help_no_version(capsys):
    parser = type("P", (), {"print_help": lambda self: print("help")})()
    args = SimpleNamespace(version=False)
    tmo_cli.print_help(args, parent_parser=parser)
    out, _ = capsys.readouterr()
    assert "help" in out


def test_list_and_select_projects_current_project(monkeypatch):
    class RM:
        def read_repo_config(self):
            return {"project_id": "p1"}

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def __iter__(self):
            return iter([{"id": "p1", "name": "proj1"}, {"id": "p2", "name": "proj2"}])

        def find_by_id(self, pid):
            if pid == "p1":
                return {"id": "p1", "name": "proj1"}
            return None

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)
    monkeypatch.setattr("builtins.input", lambda prompt="": "")

    result = tmo_cli.list_and_select_projects(
        RM(), None, as_list=False, check_config=True
    )
    assert result["id"] == "p1"


def test_list_and_select_projects_invalid_then_valid(monkeypatch):
    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def __iter__(self):
            return iter([{"id": "p1", "name": "proj1"}])

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: False)

    inputs = iter(["abc", "0"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(inputs))

    result = tmo_cli.list_and_select_projects(
        None, None, as_list=False, check_config=False
    )
    assert result["id"] == "p1"


def test_get_current_project_not_found(monkeypatch):
    class RM:
        def read_repo_config(self):
            return {"project_id": "p1"}

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_by_id(self, pid):
            return None

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)

    result = tmo_cli.get_current_project(RM(), None, check_repo_conf=False)
    assert result is None


def test_get_current_project_no_repo_config(monkeypatch):
    class RM:
        def read_repo_config(self):
            return None

    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_by_id(self, pid):
            return None

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)

    result = tmo_cli.get_current_project(RM(), None, check_repo_conf=False)
    assert result is None


def test_get_current_project_invalid_catalog():
    result = tmo_cli.get_current_project(None, None, check_repo_conf=False)
    assert result is None


def test_clone_with_project_id(monkeypatch):
    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_by_id(self, pid):
            return {
                "id": "p1",
                "name": "proj1",
                "gitRepositoryUrl": "http://git.example.com/repo",
            }

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def write_repo_config(self, config, path):
            pass

    monkeypatch.setattr("tmo.ProjectApi", PApi)

    args = SimpleNamespace(cwd=None, project_id="p1", path="/tmp/test")
    tmo_cli.clone(args, RM(), None)


def test_clone_project_not_found(monkeypatch):
    class PApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_by_id(self, pid):
            return None

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    monkeypatch.setattr(
        tmo_cli,
        "list_and_select_projects",
        lambda rm, tmo, a, b: {
            "id": "p2",
            "name": "proj2",
            "gitRepositoryUrl": "http://git.example.com/repo",
        },
    )

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def write_repo_config(self, config, path):
            pass

    args = SimpleNamespace(cwd=None, project_id="notfound", path=None)
    tmo_cli.clone(args, RM(), None)


def test_clone_no_args(monkeypatch, tmp_path):
    monkeypatch.setattr(
        tmo_cli,
        "list_and_select_projects",
        lambda rm, tmo, a, b: {
            "id": "p1",
            "name": "testproject",
            "gitRepositoryUrl": "http://git.example.com/repo",
            "branch": "main",
        },
    )

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def write_repo_config(self, config, path):
            pass

    tmo_cli.base_path = str(tmp_path)
    args = SimpleNamespace(cwd=None, project_id=None, path=None)
    tmo_cli.clone(args, RM(), None)


def test_add_task_invalid_cwd(monkeypatch):
    monkeypatch.setattr(tmo_cli, "validate_fe_tasks_cwd_valid", lambda: False)

    args = SimpleNamespace(cwd=None, template_url=None, branch=None, name=None)
    with pytest.raises(SystemExit):
        tmo_cli.add_task(args, None)


def test_add_task_no_templates(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "validate_fe_tasks_cwd_valid", lambda: True)

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def get_templates(self, entity_type=None, source_path=None):
            return {}

    args = SimpleNamespace(
        cwd=None, template_url="http://example.com", branch="main", name=None
    )
    with pytest.raises(SystemExit):
        tmo_cli.add_task(args, RM())


def test_add_task_success(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "validate_fe_tasks_cwd_valid", lambda: True)
    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "task1")

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def get_templates(self, entity_type=None, source_path=None):
            return {"task1": "/path/to/template"}

        def add_task(self, template, task_name, base_path):
            pass

    tmo_cli.base_path = str(tmp_path)
    args = SimpleNamespace(
        cwd=None, template_url="http://example.com", branch="main", name="mytask"
    )
    tmo_cli.add_task(args, RM())


def test_run_task_invalid_project(monkeypatch):
    monkeypatch.setattr(tmo_cli, "get_current_project", lambda rm, tmo, check: None)

    args = SimpleNamespace(cwd=None, connection=None, name=None, function_name=None)
    with pytest.raises(SystemExit):
        tmo_cli.run_task(args, None, None)


def test_run_task_success(monkeypatch, tmp_path):
    monkeypatch.setattr(
        tmo_cli,
        "get_current_project",
        lambda rm, tmo, check: {"id": "p1", "name": "proj1"},
    )
    monkeypatch.setattr(tmo_cli, "activate_connection", lambda args: "c1")

    class Client:
        def set_project_id(self, pid):
            pass

    class Runner:
        def __init__(self, rm):
            pass

        def run_task_local(self, base_path, task_name, func):
            return ("task1", "func1")

    import tmo

    monkeypatch.setattr(tmo, "RunTask", Runner)

    tmo_cli.base_path = str(tmp_path)
    args = SimpleNamespace(
        cwd=None, connection=None, name="task1", function_name="func1"
    )
    tmo_cli.run_task(args, None, Client())


def test_list_resources_projects(monkeypatch, capsys):
    monkeypatch.setattr(tmo_cli, "list_and_select_projects", lambda *a, **k: None)

    args = SimpleNamespace(
        cwd=None,
        projects=True,
        models=False,
        local_models=False,
        templates=False,
        datasets=False,
        connections=False,
    )

    with pytest.raises(SystemExit):
        tmo_cli.list_resources(args, None, None)


def test_list_resources_models(monkeypatch):
    monkeypatch.setattr(
        tmo_cli,
        "get_current_project",
        lambda rm, tmo: {"id": "p1", "name": "proj1"},
    )

    class Client:
        def set_project_id(self, pid):
            pass

    class MApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def __len__(self):
            return 1

        def __iter__(self):
            return iter([{"id": "m1", "name": "model1", "source": "git"}])

    import tmo

    monkeypatch.setattr(tmo, "ModelApi", MApi)

    args = SimpleNamespace(
        cwd=None,
        projects=False,
        models=True,
        local_models=False,
        templates=False,
        datasets=False,
        connections=False,
    )

    tmo_cli.list_resources(args, None, Client())


def test_list_resources_local_models(monkeypatch, capsys):
    import tmo

    monkeypatch.setattr(
        tmo.TrainModel,
        "get_model_folders",
        lambda catalog, arg: {
            "folder1": {"name": "model1", "id": "m1"},
        },
    )

    args = SimpleNamespace(
        cwd=None,
        projects=False,
        models=False,
        local_models=True,
        templates=False,
        datasets=False,
        connections=False,
    )

    tmo_cli.list_resources(args, None, None)
    out, _ = capsys.readouterr()
    assert "model1" in out


def test_list_resources_templates(monkeypatch):
    monkeypatch.setattr(
        tmo_cli,
        "get_current_project",
        lambda rm, tmo: {"id": "p1", "name": "proj1"},
    )

    class Client:
        def set_project_id(self, pid):
            pass

    class Template:
        def __init__(self):
            self.id = "t1"
            self.name = "template1"

    class TApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_all(self):
            return [Template()]

    import tmo

    monkeypatch.setattr(tmo, "DatasetTemplateApi", TApi)

    args = SimpleNamespace(
        cwd=None,
        projects=False,
        models=False,
        local_models=False,
        templates=True,
        datasets=False,
        connections=False,
    )

    tmo_cli.list_resources(args, None, Client())


def test_list_resources_datasets(monkeypatch):
    monkeypatch.setattr(
        tmo_cli,
        "get_current_project",
        lambda rm, tmo: {"id": "p1", "name": "proj1"},
    )

    class Client:
        def set_project_id(self, pid):
            pass

    class Dataset:
        def __init__(self):
            self.id = "d1"
            self.name = "dataset1"

    class DApi:
        def __init__(self, tmo_client, show_archived=False):
            pass

        def find_all(self):
            return [Dataset()]

    import tmo

    monkeypatch.setattr(tmo, "DatasetApi", DApi)

    args = SimpleNamespace(
        cwd=None,
        projects=False,
        models=False,
        local_models=False,
        templates=False,
        datasets=True,
        connections=False,
    )

    tmo_cli.list_resources(args, None, Client())


def test_list_resources_connections(monkeypatch):
    monkeypatch.setattr(tmo_cli, "list_connections", lambda args: None)

    args = SimpleNamespace(
        cwd=None,
        projects=False,
        models=False,
        local_models=False,
        templates=False,
        datasets=False,
        connections=True,
    )

    tmo_cli.list_resources(args, None, None)


def test_init_model_directory_with_existing_config(monkeypatch):
    class RM:
        def __init__(self):
            self.inited = False

        def init_model_directory(self):
            self.inited = True

        def repo_config_exists(self):
            return True

    args = SimpleNamespace(cwd=None)
    repo_manager = RM()
    tmo_cli.init_model_directory(args, repo_manager, None)
    assert repo_manager.inited is True


def test_add_model_with_prompts(monkeypatch, tmp_path):
    monkeypatch.setattr(tmo_cli, "validate_model_catalog_cwd_valid", lambda: True)

    inputs = iter(["http://example.com", "main", "model1", "desc"])
    monkeypatch.setattr("builtins.input", lambda prompt="": next(inputs))
    monkeypatch.setattr(
        tmo_cli,
        "input_select",
        lambda *a, **k: "python" if "language" in str(a) else "Template 1 (t1)",
    )

    class RM:
        def clone_repository(self, url, path, branch):
            pass

        def get_templates(self, entity_type=None, source_path=None):
            return {
                "python": {
                    "t1": ["Template 1", "/path/to/template"],
                }
            }

        def add_model(self, model_id, model_name, model_desc, template, base_path):
            pass

    tmo_cli.base_path = str(tmp_path)
    args = SimpleNamespace(cwd=None, template_url=None, branch=None)
    tmo_cli.add_model(args, RM())


def test_input_string_password_with_is_called_from_test_true(monkeypatch):
    monkeypatch.setattr("builtins.input", lambda prompt="": "secret")
    result = tmo_cli.input_string("test", password=True, is_called_from_test=True)
    assert result == "secret"


def test_input_select_with_label(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda prompt="": "0")
    result = tmo_cli.input_select("item", ["a", "b"], label="Choose one:")
    assert result == "a"
    out, _ = capsys.readouterr()
    assert "Choose one:" in out


def test_yes_or_no_case_insensitive(monkeypatch):
    monkeypatch.setattr("builtins.input", lambda prompt="": "Y")
    assert tmo_cli.yes_or_no("question") is True

    monkeypatch.setattr("builtins.input", lambda prompt="": "N")
    assert tmo_cli.yes_or_no("question") is False


def test_clone_with_project_found(monkeypatch, tmp_path):
    args = type("A", (), {"cwd": None, "project_id": "p1", "path": None})()

    class RM:
        def __init__(self):
            self.cloned = False
            self.config = None

        def clone_repository(self, git, path, branch):
            self.cloned = True

        def write_repo_config(self, config, path=None):
            self.config = (config, path)

    class PApi:
        def __init__(self, tmo_client=None, show_archived=False):
            pass

        def find_by_id(self, pid):
            return {
                "id": pid,
                "name": "proj",
                "gitRepositoryUrl": "git@x",
                "branch": "main",
            }

    monkeypatch.setattr("tmo.ProjectApi", PApi)
    rm = RM()
    tmo_cli.clone(args, rm, None)
    assert rm.cloned is True
    assert rm.config[0]["project_id"] == "p1"


def test_list_connections_prints_when_connections_present(tmp_path, capsys):
    cfg = tmp_path / "cfg"
    tmo_cli.config_dir = str(cfg)
    cfg.mkdir(parents=True)
    con = {
        "connections": [{
            "id": "c1",
            "name": "n",
            "username": "u",
            "password": "p",
            "host": "h",
            "logmech": "TDNEGO",
            "database": "db",
        }]
    }
    yaml.safe_dump(con, open(cfg / "connections.yaml", "w+"))
    args = type("A", (), {"cwd": None})()
    tmo_cli.list_connections(args)
    out, _ = capsys.readouterr()
    assert "List of local connections" in out or "Name: n" in out


def test_add_connections_invalid_args_calls_help_and_exits(monkeypatch):
    args = type(
        "A",
        (),
        {
            "cwd": None,
            "name": None,
            "username": "u",
            "password": None,
            "host": None,
            "database": None,
            "val_db": None,
            "byom_db": None,
            "logmech": None,
            "parent_parser": type("P", (), {"print_help": lambda self: None})(),
        },
    )()
    with pytest.raises(SystemExit):
        tmo_cli.add_connections(args)


def _write_tmp_json(tmp_path, content):
    p = tmp_path / "tmp.json"
    p.write_text(content)
    return str(p)


def test_run_model_calls_train_evaluate_score(tmp_path, monkeypatch):
    import tmo

    model_id = "m1"
    available = {0: {"id": model_id, "name": "M"}}
    monkeypatch.setattr(tmo.TrainModel, "get_model_ids", lambda catalog, arg: available)
    monkeypatch.setattr(
        tmo_cli, "get_current_project", lambda rm, tc, check: {"id": "p1"}
    )

    class RepoManager:
        pass

    class Client:
        def set_project_id(self, pid):
            pass

    # prepare a local dataset file to bypass selection
    dataset_path = _write_tmp_json(tmp_path, '{"data": []}')

    # Train mode
    args = type(
        "A",
        (),
        {
            "cwd": None,
            "model_id": model_id,
            "mode": "train",
            "local_dataset": dataset_path,
            "local_dataset_template": None,
            "dataset_id": None,
            "dataset_template_id": None,
            "connection": None,
        },
    )()
    called = {}

    class FakeTrainer:
        @staticmethod
        def get_model_ids(catalog, arg):
            return available

        def __init__(self, repo_manager):
            pass

        def train_model_local(self, mid, rendered_dataset=None, base_path=None):
            called["train"] = (mid, rendered_dataset is not None)

    monkeypatch.setattr(tmo, "TrainModel", FakeTrainer)
    tmo_cli.run_model(args, RepoManager(), Client())
    assert called.get("train")[0] == model_id

    # Evaluate mode
    called.clear()
    args.mode = "evaluate"

    class FakeEvaluator:
        def __init__(self, repo_manager):
            pass

        def evaluate_model_local(self, mid, rendered_dataset=None, base_path=None):
            called["evaluate"] = (mid, rendered_dataset is not None)

    monkeypatch.setattr(tmo, "EvaluateModel", FakeEvaluator)
    tmo_cli.run_model(args, RepoManager(), Client())
    assert called.get("evaluate")[0] == model_id

    # Score mode uses local dataset template bypass
    called.clear()
    args.mode = "score"
    args.local_dataset = None
    tmp_template = _write_tmp_json(tmp_path, '{"template": true}')
    args.local_dataset_template = tmp_template

    class FakeScorer:
        def __init__(self, repo_manager):
            pass

        def batch_score_model_local(self, mid, rendered_dataset=None, base_path=None):
            called["score"] = (mid, rendered_dataset is not None)

    monkeypatch.setattr(tmo, "ScoreModel", FakeScorer)
    tmo_cli.run_model(args, RepoManager(), Client())
    assert called.get("score")[0] == model_id


def test_import_stats_saves_feature_stats(tmp_path, monkeypatch):
    import tmo

    stats_json = {
        "features": {
            "age": {"type": "continuous", "edges": [1, 2]},
            "flag": {"type": "categorical", "frequencies": ["0", "1"]},
        }
    }
    p = tmp_path / "stats.json"
    p.write_text(json.dumps(stats_json))

    calls = []
    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)
    monkeypatch.setattr(
        "tmo.stats.store.save_feature_stats", lambda *a, **k: calls.append((a, k))
    )

    args = type(
        "A",
        (),
        {
            "cwd": None,
            "show_example": False,
            "statistics_file": str(p),
            "metadata_table": "mt",
        },
    )()
    tmo_cli.import_stats(args)
    # should have saved categorical and continuous
    assert any(call[0][1] == "categorical" for call in calls)
    assert any(call[0][1] == "continuous" for call in calls)


def test_compute_stats_calls_store(tmp_path, monkeypatch):
    import tmo

    monkeypatch.setattr(tmo_cli, "activate_connection", lambda ns: None)
    monkeypatch.setattr(tmo, "tmo_create_context", lambda: None)

    class FakeDF:
        pass

    # mock teradataml.DataFrame.from_query
    import teradataml

    monkeypatch.setattr(
        teradataml.DataFrame, "from_query", classmethod(lambda cls, q: FakeDF())
    )

    # categorical
    monkeypatch.setattr(
        "tmo.stats.stats.compute_categorical_stats",
        lambda df, cols, temp_db=None: {"a": "cat"},
    )
    saved = []
    monkeypatch.setattr(
        "tmo.stats.store.save_feature_stats", lambda *a, **k: saved.append((a, k))
    )
    args = type(
        "A",
        (),
        {
            "cwd": None,
            "source_table": "t",
            "columns": "a",
            "feature_type": "categorical",
            "metadata_table": "mt",
            "temp_view_database": None,
        },
    )()
    tmo_cli.compute_stats(args)
    assert any(call[1].get("feature_type") == "categorical" for call in saved)

    # continuous
    saved.clear()
    monkeypatch.setattr(
        "tmo.stats.stats.compute_continuous_stats",
        lambda df, cols, temp_db=None: {"a": "cont"},
    )
    args.feature_type = "continuous"
    tmo_cli.compute_stats(args)
    assert any(call[1].get("feature_type") == "continuous" for call in saved)


def test_check_connection_exists_found():
    """Test _check_connection_exists when connection is found."""
    connections = [
        {"id": "c1", "name": "conn1"},
        {"id": "c2", "name": "conn2"},
    ]
    assert tmo_cli._check_connection_exists("c1", connections) is True
    assert tmo_cli._check_connection_exists("c3", connections) is False


def test_check_connection_exists_empty_list():
    """Test _check_connection_exists with empty list."""
    assert tmo_cli._check_connection_exists("c1", []) is False


def test_select_connection_from_list_single(capsys):
    """Test _select_connection_from_list with single connection."""
    connections = [{"id": "c1", "name": "conn1"}]
    result = tmo_cli._select_connection_from_list(connections)
    assert result == "c1"
    out, _ = capsys.readouterr()
    assert "Automatic connection selection" in out


def test_select_connection_from_list_multiple(monkeypatch):
    """Test _select_connection_from_list with multiple connections."""
    connections = [
        {"id": "c1", "name": "conn1"},
        {"id": "c2", "name": "conn2"},
    ]
    monkeypatch.setattr(tmo_cli, "input_select", lambda *a, **k: "conn2")
    result = tmo_cli._select_connection_from_list(connections)
    assert result == "c2"


def test_set_connection_env_vars_success():
    """Test _set_connection_env_vars sets environment variables correctly."""
    connections = [{
        "id": "c1",
        "name": "conn1",
        "username": "user1",
        "password": "pass1",
        "host": "host1",
        "logmech": "TDNEGO",
        "database": "db1",
        "val_db": "VAL",
        "ml_db": "MLDB",
    }]
    result = tmo_cli._set_connection_env_vars("c1", connections)
    assert result is True
    assert os.environ.get("VMO_CONN_USERNAME") == "user1"
    assert os.environ.get("VMO_CONN_HOST") == "host1"


def test_set_connection_env_vars_not_found():
    """Test _set_connection_env_vars when connection not found."""
    connections = [{"id": "c1", "name": "conn1"}]
    result = tmo_cli._set_connection_env_vars("c2", connections)
    assert result is False


def test_get_connections_list_with_connections():
    """Test _get_connections_list with valid connections."""
    connections_dict = {
        "connections": [
            {"id": "c1", "name": "conn1"},
            {"id": "c2", "name": "conn2"},
        ]
    }
    result = tmo_cli._get_connections_list(connections_dict)
    assert len(result) == 2
    assert result[0]["id"] == "c1"


def test_get_connections_list_empty():
    """Test _get_connections_list with empty or missing connections."""
    assert tmo_cli._get_connections_list({}) == []
    assert tmo_cli._get_connections_list({"connections": []}) == []


def test_determine_connection_id_from_args():
    """Test _determine_connection_id uses args.connection when available."""
    args = SimpleNamespace(connection="c1")
    kwargs = {}
    connections = [{"id": "c1", "name": "conn1"}]
    result = tmo_cli._determine_connection_id(args, kwargs, connections)
    assert result == "c1"


def test_determine_connection_id_from_kwargs():
    """Test _determine_connection_id uses kwargs when args doesn't have connection."""
    args = SimpleNamespace(connection=None)
    kwargs = {"connection": "c2"}
    connections = [{"id": "c2", "name": "conn2"}]
    result = tmo_cli._determine_connection_id(args, kwargs, connections)
    assert result == "c2"


def test_determine_connection_id_select_from_list(monkeypatch):
    """Test _determine_connection_id selects from list when no args/kwargs."""
    args = SimpleNamespace(connection=None)
    kwargs = {}
    connections = [{"id": "c1", "name": "conn1"}]
    result = tmo_cli._determine_connection_id(args, kwargs, connections)
    assert result == "c1"


def test_print_projects_list_with_projects(capsys):
    """Test _print_projects_list prints projects correctly."""
    projects = [
        {"id": "p1", "name": "Project 1"},
        {"id": "p2", "name": "Project 2"},
    ]
    tmo_cli._print_projects_list(projects, as_list=False)
    out, _ = capsys.readouterr()
    assert "Available projects:" in out
    assert "p1" in out
    assert "Project 1" in out


def test_print_projects_list_empty(capsys):
    """Test _print_projects_list with empty list."""
    tmo_cli._print_projects_list([], as_list=True)
    out, _ = capsys.readouterr()
    assert "List of projects:" in out
    assert "No projects were found" in out


def test_find_current_project_index_found():
    """Test _find_current_project_index when project is found."""
    projects = [
        {"id": "p1", "name": "Project 1"},
        {"id": "p2", "name": "Project 2"},
    ]
    current_project = {"id": "p2"}
    result = tmo_cli._find_current_project_index(projects, current_project)
    assert result == 1


def test_find_current_project_index_not_found():
    """Test _find_current_project_index when project is not found."""
    projects = [{"id": "p1", "name": "Project 1"}]
    current_project = {"id": "p3"}
    result = tmo_cli._find_current_project_index(projects, current_project)
    assert result == "none"


def test_find_current_project_index_none():
    """Test _find_current_project_index with None current project."""
    projects = [{"id": "p1", "name": "Project 1"}]
    result = tmo_cli._find_current_project_index(projects, None)
    assert result == "none"


def test_validate_project_selection_valid():
    """Test _validate_project_selection with valid selection."""
    projects = [{"id": "p1"}, {"id": "p2"}]
    assert tmo_cli._validate_project_selection("0", projects, "none") is True
    assert tmo_cli._validate_project_selection("1", projects, "none") is True
    assert tmo_cli._validate_project_selection("", projects, 0) is True


def test_validate_project_selection_invalid():
    """Test _validate_project_selection with invalid selection."""
    projects = [{"id": "p1"}, {"id": "p2"}]
    assert tmo_cli._validate_project_selection("2", projects, "none") is False
    assert tmo_cli._validate_project_selection("abc", projects, "none") is False
    assert tmo_cli._validate_project_selection("", projects, "none") is False


def test_check_if_any_resource_selected_true():
    """Test _check_if_any_resource_selected returns True when resource selected."""
    args = SimpleNamespace(
        projects=True,
        models=False,
        local_models=False,
        templates=False,
        datasets=False,
        connections=False,
    )
    assert tmo_cli._check_if_any_resource_selected(args) is True


def test_check_if_any_resource_selected_false():
    """Test _check_if_any_resource_selected returns False when none selected."""
    args = SimpleNamespace(
        projects=False,
        models=False,
        local_models=False,
        templates=False,
        datasets=False,
        connections=False,
    )
    assert tmo_cli._check_if_any_resource_selected(args) is False


def test_handle_invalid_grant_error_token_not_active(monkeypatch):
    """Test _handle_invalid_grant_error with 'Token is not active' error."""

    class FakeError:
        description = "Token is not active"

    args = SimpleNamespace(debug=False)
    remove_called = {"called": False}

    def fake_remove():
        remove_called["called"] = True

    with pytest.raises(SystemExit) as exc:
        tmo_cli._handle_invalid_grant_error(FakeError(), args, fake_remove)

    assert exc.value.code == 1
    assert remove_called["called"] is True


def test_handle_invalid_grant_error_session_not_active(monkeypatch):
    """Test _handle_invalid_grant_error with 'Session not active' error."""

    class FakeError:
        description = "Session not active"

    args = SimpleNamespace(debug=False)
    remove_called = {"called": False}

    def fake_remove():
        remove_called["called"] = True

    with pytest.raises(SystemExit) as exc:
        tmo_cli._handle_invalid_grant_error(FakeError(), args, fake_remove)

    assert exc.value.code == 1
    assert remove_called["called"] is True


def test_handle_invalid_grant_error_other_error(monkeypatch):
    """Test _handle_invalid_grant_error with other error types."""

    class FakeError:
        description = "Some other error"

    args = SimpleNamespace(debug=False)

    def fake_remove():
        pass

    # Should call handle_generic_error which exits with code 1
    with pytest.raises(SystemExit) as exc:
        tmo_cli._handle_invalid_grant_error(FakeError(), args, fake_remove)

    assert exc.value.code == 1


def test_load_connections_from_file_success(tmp_path):
    """Test _load_connections_from_file loads file successfully."""
    tmp = tmp_path / "cfg"
    tmo_cli.config_dir = str(tmp)
    tmp.mkdir(parents=True)
    connections_data = {"connections": [{"id": "c1", "name": "conn1"}]}
    yaml.safe_dump(
        connections_data,
        open(Path(tmo_cli.config_dir) / "connections.yaml", "w+"),
    )
    result = tmo_cli._load_connections_from_file()
    assert "connections" in result
    assert len(result["connections"]) == 1


def test_load_connections_from_file_not_found(tmp_path):
    """Test _load_connections_from_file when file doesn't exist."""
    tmp = tmp_path / "cfg_nonexistent"
    tmo_cli.config_dir = str(tmp)
    with pytest.raises(SystemExit) as exc:
        tmo_cli._load_connections_from_file()
    assert exc.value.code == 1



from enum import Enum


class JobAutomationMode(Enum):
    TRAIN = "training"
    EVALUATE = "evaluation"
    SCORE = "scoring"
    DEPLOY = "deployment"


class JobRunnerMode(Enum):
    TRAIN = "Train"
    EVALUATE = "Evaluate"
    SCORE = "Score"


class JobRunnerEngine(Enum):
    PYTHON = "python"
    R = "R"
    SQL = "sql"



__version__ = "7.2.3"

# Import client
from tmo.api_client import TmoClient

# Import APIs from api package
from tmo.api.api_iterator import ApiIterator
from tmo.api.dataset_api import DatasetApi
from tmo.api.dataset_connection_api import DatasetConnectionApi
from tmo.api.dataset_template_api import DatasetTemplateApi
from tmo.api.deployment_api import DeploymentApi
from tmo.api.feature_engineering_api import FeatureEngineeringApi
from tmo.api.job_api import JobApi
from tmo.api.job_event_api import JobEventApi
from tmo.api.message_api import MessageApi
from tmo.api.model_api import ModelApi
from tmo.api.project_api import ProjectApi
from tmo.api.trained_model_api import TrainedModelApi
from tmo.api.trained_model_artefacts_api import TrainedModelArtefactsApi
from tmo.api.trained_model_event_api import TrainedModelEventApi
from tmo.api.user_attributes_api import UserAttributesApi

# import cli package
from tmo.cli.base_model import BaseModel
from tmo.cli.base_task import BaseTask
from tmo.cli.evaluate_model import EvaluateModel
from tmo.cli.repo_manager import RepoManager
from tmo.cli.run_task import RunTask
from tmo.cli.score_model import ScoreModel
from tmo.cli.train_model import TrainModel

# import context package
from tmo.context.model_context import *

# import crypto package
from tmo.crypto import crypto

# import decorators package
from tmo.decorators import *

# import stats package
from tmo.stats.stats import *
from tmo.stats.store import *

# import types package
from tmo.types import *

# import util package
from tmo.util import *



from __future__ import absolute_import

from tmo.api.iterator_base_api import IteratorBaseApi


class JobEventApi(IteratorBaseApi):

    path = "/api/jobEvents"
    type = "JOB_EVENT"

    def _get_header_params(self):
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all job events

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all job events
        """
        query_vars = ["projection", "page", "sort", "size", "sort"]
        query_vals = [projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_id(self, job_event_id: str, projection: str = None):
        """
        returns a job event

        Parameters:
           job_event_id (str): job event id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): job event
        """
        query_vars = ["id", "projection"]
        query_vals = [job_event_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findById",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_job_id(
        self,
        job_id: str,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns events of a job

        Parameters:
           job_id (str): job id(uuid) to find events
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): job events
        """
        query_vars = ["jobId", "projection", "page", "sort", "size", "sort"]
        query_vals = [job_id, projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByJobId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_status(
        self,
        status: list[str],
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns job by status

        Parameters:
           status (str): job status
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): job events
        """
        query_vars = ["status", "projection", "page", "sort", "size"]
        query_vals = [status, projection, page, sort, size]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByStatusIn",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_job_between_intervals(
        self, start_time: str, end_time: str, projection: str = None
    ):
        """
        returns job events within the timestamp range

        Parameters:
           start_time (str): start time e.g: 2020-03-18T09:05:03.569Z
           end_time (str): end time e.g: 2020-03-24T09:05:03.569Z
           projection (str): projection type

        Returns:
            (dict): job events
        """
        query_vars = ["startAt", "endAt", "projection"]
        query_vals = [start_time, end_time, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByCreatedAtBetween",
            header_params=self._get_header_params(),
            query_params=query_params,
        )



import importlib
import json
import logging
import os
import shutil
import subprocess
import sys

from tmo.cli.base_model import BaseModel
from tmo.context.model_context import ModelContext, DatasetInfo
from tmo.types.job import JobRunnerMode, JobRunnerEngine, JobAutomationMode

ARTIFACTS_PATH = "./artifacts"
MODELS_PATH = "./models"


class TrainModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def train_model_local(self, model_id: str, rendered_dataset: dict, base_path: str):

        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = f"{base_path}model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(f"model directory {model_definitions_path} does not exist")

        model_artefacts_path = f".artefacts/{model_id}/"  # noqa
        model_artefacts_abs_path = os.path.abspath(model_artefacts_path)
        model_artefacts_output_path = os.path.join(model_artefacts_path, "output/")
        if os.path.exists(model_artefacts_path):
            self.logger.debug(
                f"Cleaning local model artefact path: {model_artefacts_path}"
            )
            shutil.rmtree(model_artefacts_path)

        os.makedirs(model_artefacts_output_path)

        try:
            if os.path.exists(ARTIFACTS_PATH):
                os.remove(ARTIFACTS_PATH)

            if os.name == "nt":
                subprocess.run(
                    [
                        "cmd",
                        "/c",
                        "mklink",
                        "/J",
                        "artifacts",
                        model_artefacts_abs_path,
                    ],
                    check=True,
                    capture_output=True,
                    text=True,
                )
            else:
                os.symlink(
                    model_artefacts_path, ARTIFACTS_PATH, target_is_directory=True
                )

            model_dir = model_definitions_path + BaseModel.get_model_folder(  # noqa
                model_definitions_path, model_id
            )

            with open(f"{model_dir}/model.json", "r") as f:
                model_definition = json.load(f)

            with open(f"{model_dir}/config.json", "r") as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kwargs = self._BaseModel__get_model_varargs(model_id)  # noqa

            context = ModelContext(
                dataset_info=DatasetInfo.from_dict(rendered_dataset),
                hyperparams=model_conf["hyperParameters"],
                artifact_output_path="artifacts/output",
                **cli_model_kwargs,
            )

            engine = JobRunnerEngine(
                self._BaseModel__get_engine(  # noqa
                    model_definition, JobAutomationMode.TRAIN
                )
            )
            if engine == JobRunnerEngine.PYTHON:
                sys.path.append(model_dir)
                training = importlib.import_module(".training", package="model_modules")
                training.train(
                    context=context,
                    data_conf=rendered_dataset,
                    model_conf=model_conf,
                    **cli_model_kwargs,
                )

            elif engine == JobRunnerEngine.SQL:
                self.__train_sql(
                    context, model_dir, rendered_dataset, model_conf, **cli_model_kwargs
                )

            elif engine == JobRunnerEngine.R:
                self._BaseModel__run_r_model(  # noqa
                    model_id,
                    model_dir,
                    rendered_dataset,
                    JobRunnerMode.TRAIN.value.lower(),
                )

            else:
                raise ValueError(f"Unsupported engine: {engine}")

            self.logger.info(
                f"Artefacts can be found in: {model_artefacts_output_path}"
            )
            self.__cleanup()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                f"pip install -r {model_dir}/model_modules/requirements.txt"
            )
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    @staticmethod
    def __cleanup():
        if os.path.exists(ARTIFACTS_PATH):
            os.remove(ARTIFACTS_PATH)
        if os.path.exists(MODELS_PATH):
            os.remove(MODELS_PATH)

    def __train_sql(
        self,
        context: ModelContext,
        model_dir: str,
        data_conf: dict,
        model_conf: dict,
        **kwargs,
    ):
        from tmo.util import tmo_create_context
        from teradataml import remove_context

        self.logger.info("Starting training...")

        tmo_create_context()  # noqa

        sql_file = f"{model_dir}/model_modules/training.sql"
        jinja_ctx = {
            "context": context,
            "data_conf": data_conf,
            "model_conf": model_conf,
            "model_table": kwargs.get("model_table"),
            "model_version": kwargs.get("model_version"),
            "model_id": kwargs.get("model_id"),
        }

        self._BaseModel__execute_sql_script(sql_file, jinja_ctx)  # noqa

        remove_context()

        self.logger.info("Finished training")

        self.logger.info("Saved trained model")

        remove_context()



import os

import yaml

REQUIREMENTS_FILE = "/requirements.txt"
TASK_PY_FILE = "/task.py"


class BaseTask(object):

    def __init__(self, repo_manager):
        self.repo_manager = repo_manager

    def __get_task_varargs(self):  # noqa # NOSONAR
        return {"project_id": self.__get_project_id(), "job_id": "cli"}

    def __get_project_id(self):
        self.repo_manager.repo_config_rename(self.repo_manager.base_path)
        path = os.path.join(self.repo_manager.base_path, ".tmo/config.yaml")
        with open(path, "r") as handle:
            return yaml.safe_load(handle)["project_id"]

    @staticmethod
    def get_task_folders(task_path: str) -> list[str]:
        catalog = []
        for task_folder in os.listdir(task_path):
            if os.path.exists(
                task_path + task_folder + REQUIREMENTS_FILE
            ) and os.path.exists(task_path + task_folder + TASK_PY_FILE):
                catalog.append(task_folder)

        return catalog



from __future__ import absolute_import

import logging
import uuid
from typing import Union

from teradataml import DataFrame

from tmo.api.dataset_template_api import (
    DatasetTemplateApi,
    FeaturesEntityTargets,
    Metadata,
    Predictions,
    Variable,
)
from tmo.api.iterator_base_api import IteratorBaseApi
from tmo.types.dataset import Dataset, Scope
from tmo.types.dataset_metadata import TypeEnum
from tmo.types.exceptions import EntityNotFoundError

logger = logging.getLogger(__name__)


class DatasetApi(IteratorBaseApi):
    path = "/api/datasets"
    type = "DATASET"

    def create(
        self,
        dataset_template_id: uuid.UUID,
        name: str,
        description: str,
        scope: Scope,
        sql: dict[str, str] = None,
        tables: dict = None,
    ) -> Dataset:
        """
        Creates a dataset.

        Parameters:
            dataset_template_id (UUID): The dataset template id.
            name (str): The name of the dataset.
            description (str): The description of the dataset.
            scope (str): The scope of the dataset can be 'train' or 'evaluate'
            sql (str, optional): SQL query used to select entity sample and target. If None, this query will be automatically created using the dataset template tables.
                - entityAndTargets: SQL query for extracting entity and targets
                - predictions: SQL query for extracting entity and targets for predictions
            tables (dict, optional): The tables to be used in the dataset. If None, the tables will be automatically created using the dataset template name as a base.
                - data: Contains entity, features and targets.
                - entityTarget: Stores entity and target. If not provided these will be extracted from 'dataTable'.
                - predictions: Contains entity and target for predictions. If not provided, it will be created using the dataset template name as a base.

        Returns:
            (Dataset): dataset

        Example:
            ```python
            from tmo import TmoClient

            vmoClient = TmoClient()

            train_dataset = (
                vmoClient
                .datasets()
                .create(
                    dataset_template_id="1a71337c-8b6f-4500-a129-ef8036578c81",
                    name="Training",
                    description="dataset description",
                    scope="train",
                )
            )
            ```
        """

        sql = sql or {}
        tables = tables or {}

        dataset_template = DatasetTemplateApi(self.tmo_client).find_by_id(
            dataset_template_id
        )

        if dataset_template is None:
            raise EntityNotFoundError(
                f"Dataset template with id {dataset_template_id} not found"
            )

        dataset = Dataset(
            dataset_template_id=dataset_template_id,
            name=name,
            description=description,
            scope=scope,
        )

        entity_and_target = dataset_template.metadata.entity_and_targets.variables
        database = dataset_template.metadata.predictions.database
        dataset_template_metadata = dataset_template.metadata
        features = dataset_template.metadata.features.variables
        features_sql = dataset_template.metadata.features.sql

        entity_and_target_sql = sql.get(
            "entityAndTargets", dataset_template_metadata.entity_and_targets.sql
        )
        predictions_entity_sql = sql.get(
            "predictions", dataset_template_metadata.predictions.entity_sql
        )

        data_table = tables.get(
            "data", f"{dataset_template.name.replace(' ', '_')}_data"
        )
        entity_target_table = tables.get("entityTarget", data_table)
        predictions_table = tables.get(
            "predictions", dataset_template_metadata.predictions.table
        )

        entity_columns_objects = [
            Variable()
            .set_name(col.name)
            .set_data_type(col.data_type)
            .set_type(col.type)
            .set_entity_id(True)
            .set_selected(False)
            for col in entity_and_target
            if col.type == TypeEnum.ENTITY
        ]

        target_columns_objects = [
            Variable()
            .set_name(col.name)
            .set_data_type(col.data_type)
            .set_type(col.type)
            .set_entity_id(False)
            .set_selected(True)
            for col in entity_and_target
            if col.type == TypeEnum.TARGET
        ]

        entity = entity_columns_objects[0].name

        features = (
            FeaturesEntityTargets()
            .set_sql(features_sql)
            .set_entity(entity)
            .set_columns(features)
        )

        entity_columns = [col.name for col in entity_columns_objects]
        target_columns = [col.name for col in target_columns_objects]

        # Create entity targets object with custom SQL if available
        entity_targets = FeaturesEntityTargets().set_entity(entity)
        if sql and "target_entity" in sql:
            logger.debug(f"Using custom SQL for target_entity: {sql['target_entity']}")
            entity_targets.sql(sql["target_entity"])
        else:
            query = (
                f"SELECT {entity_columns[0]}, {', '.join(target_columns)} FROM"
                f" {entity_target_table}"
            )
            logger.debug(
                f"Using data table for target_entity SQL: {entity_target_table}"
            )
            entity_targets.set_sql(query)
        entity_targets.set_columns(entity_columns_objects + target_columns_objects)

        predictions = (
            Predictions()
            .set_database(database)
            .set_entity_sql(predictions_entity_sql)
            .set_table(predictions_table)
        )

        dataset.metadata = (
            Metadata()
            .set_type("CatalogBody")
            .set_features(features)
            .set_entity_and_targets(entity_targets)
            .set_predictions(predictions)
        )

        dataset_request = {
            "datasetTemplateId": str(dataset.dataset_template_id),
            "name": dataset.name,
            "description": dataset.description,
            "scope": dataset.scope.value,
            "metadata": {
                "features": {
                    "sql": features_sql,
                    "entity": entity,
                    "variables": [
                        {
                            "name": col.name,
                            "type": col.type.value,
                            "dataType": col.data_type.value,
                            "selected": col.selected,
                            "entityId": str(col.entity_id),
                        }
                        for col in features.variables
                    ],
                },
                "entityAndTargets": {
                    "entity": entity,
                    "sql": entity_and_target_sql,
                    "variables": [
                        {
                            "name": col.name,
                            "dataType": col.data_type.value,
                            "type": col.type.value,
                        }
                        for col in entity_and_target
                    ],
                },
                "predictions": {
                    "database": database,
                    "entitySql": predictions_entity_sql,
                    "table": predictions_table,
                },
                "type": dataset_template.metadata.type.value,
            },
            "catalogType": dataset_template.catalog_type.value,
        }

        response = self.tmo_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset_request,
        )

        dataset.id = uuid.UUID(response.get("id"))

        logger.debug("Dataset created successfully.")

        return dataset

    def save(self, dataset: dict[str, str]):
        """
        register a dataset

        Parameters:
           dataset (dict): dataset to register

        Returns:
            (dict): dataset
        """
        return self.tmo_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset,
        )

    def render(self, id: str | uuid.UUID) -> dict:
        """
        returns a rendered dataset

        Parameters:
           id (str): dataset id

        Returns:
            (dict): rendered dataset
        """

        return self.tmo_client.get_request(
            path=f"{self.path}/{str(id)}/render",
            header_params=self._get_header_params(),
            query_params={},
        )

    def find_by_name_like(
        self, name: str, projection: str = None, return_dataframe: bool = False
    ) -> DataFrame | list[Dataset]:
        """
        Returns datasets matching the name as a combined DataFrame or list.

        Parameters:
            name (str): dataset name(string) to match
            projection (str): projection type
            return_dataframe (bool): if True, returns combined DataFrame; if False, returns list of Dataset objects

        Returns:
            DataFrame or list[Dataset]: combined DataFrame of all datasets or list of Dataset objects
        """
        query_vars = ["name", "projection"]
        query_vals = [name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        response = self.tmo_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        return self._process_datasets_response(
            response, return_dataframe, f" for: {name}"
        )

    def find_by_dataset_template_id(
        self,
        dataset_template_id: Union[str, uuid.UUID],
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
        return_dataframe: bool = False,
    ) -> DataFrame | list[Dataset] | None:
        """
        Returns all datasets of a project by dataset template id as a combined DataFrame or list.

        Parameters:
            dataset_template_id (str|UUID): dataset template id
            archived (bool): archived or not (default False)
            projection (str): projection type
            page (int): page number
            size (int): number of records in a page
            sort (str): column name and sorting order
                e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order
            return_dataframe (bool): if True, returns combined DataFrame; if False, returns list of Dataset objects

        Returns:
            DataFrame or list[Dataset] or None: combined DataFrame of all datasets or list of Dataset objects
        """
        # Validate and convert UUID
        validated_id = self._validate_and_convert_uuid(
            dataset_template_id, "dataset_template_id"
        )
        if validated_id is None:
            return None if return_dataframe else []

        # Build query parameters
        query_vars = [
            "datasetTemplateId",
            "archived",
            "projection",
            "page",
            "size",
            "sort",
        ]
        query_vals = [validated_id, archived, projection, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        # Execute API request
        response = self.tmo_client.get_request(
            path=f"{self.path}/search/findByDatasetTemplateId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        # Process and return response
        return self._process_datasets_response(
            response, return_dataframe, f" for template id: {validated_id}"
        )

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
        return_dataframe: bool = False,
    ):
        """
        Returns all datasets.

        Parameters:
            projection (str): projection type
            page (int): page number
            size (int): number of records in a page
            sort (str): column name and sorting order
            return_dataframe (bool): if True, returns DataFrame; if False, returns list of Dataset objects

        Returns:
            DataFrame or list[Dataset]: combined DataFrame of all datasets or list of Dataset objects
        """
        query_vars = ["projection", "page", "size", "sort"]
        query_vals = [projection, page, size, sort]
        built_query_params = self.generate_params(query_vars, query_vals)

        response = self.tmo_client.get_request(
            path=f"{self.path}",
            header_params=self._get_header_params(),
            query_params=built_query_params,
        )

        return self._process_datasets_response(response, return_dataframe)

    def _get_header_params(self):
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    def _process_datasets_response(
        self, response: dict, return_dataframe: bool, context: str = ""
    ) -> DataFrame | list[Dataset] | None:
        """
        Helper method to process API response containing datasets.
        Reduces cognitive complexity by extracting common processing logic.

        Parameters:
            response (dict): API response containing datasets
            return_dataframe (bool): if True, returns DataFrame; if False, returns list
            context (str): context message for logging (e.g., "for name: xyz")

        Returns:
            DataFrame or list[Dataset] or None: processed datasets
        """
        try:
            datasets_data = response.get("_embedded", {}).get("datasets", [])

            if not datasets_data:
                logger.info(f"No datasets found{context}")
                return None if return_dataframe else []

            datasets = []
            for dataset_data in datasets_data:
                dataset = self._parse_dataset_from_response(dataset_data)
                if dataset:
                    datasets.append(dataset)

            return (
                self._datasets_to_dataframe(datasets) if return_dataframe else datasets
            )

        except Exception as e:
            logger.error(f"Error while parsing dataset response: {str(e)}")
            return None if return_dataframe else []

    @staticmethod
    def _validate_and_convert_uuid(
        value: Union[str, uuid.UUID], param_name: str = "id"
    ) -> uuid.UUID | None:
        """
        Helper method to validate and convert a value to UUID.

        Parameters:
            value: value to convert (str or UUID)
            param_name: parameter name for logging

        Returns:
            UUID or None: converted UUID or None if invalid
        """
        if isinstance(value, uuid.UUID):
            return value

        try:
            return uuid.UUID(value)
        except ValueError:
            logger.error(f"Invalid UUID format for {param_name}: {value}")
            return None

    @staticmethod
    def _parse_dataset_from_response(dataset_data: dict) -> Dataset | None:
        """Helper method to parse a single dataset from API response"""
        try:
            dataset = Dataset(
                dataset_template_id=uuid.UUID(dataset_data.get("datasetTemplateId")),
                name=dataset_data.get("name"),
                description=dataset_data.get("description"),
                scope=Scope(dataset_data.get("scope")),
            )

            dataset.id = dataset_data.get("id")

            # get metadata
            metadata = dataset_data.get("metadata")
            if not metadata:
                return dataset

            features_data = metadata.get("features", {})
            entity_targets_data = metadata.get("entityAndTargets", {})
            predictions_data = metadata.get("predictions", {})

            # Extract entity
            entity = features_data.get("entity")
            dataset.entity = entity

            # Extract targets
            if "variables" in entity_targets_data:
                target_variables = []
                for variable in entity_targets_data.get("variables", []):
                    if variable.get("type") == "target":
                        target_variables.append(variable.get("name"))
                dataset.target = target_variables

            # Set up the metadata structure
            dataset.metadata = Metadata()
            dataset.metadata.features = FeaturesEntityTargets()
            dataset.metadata.entityAndTargets = FeaturesEntityTargets()
            dataset.metadata.predictions = Predictions()
            dataset.metadata.type = metadata.get("type")

            # Parse entity target columns
            entity_target_columns = []  # noqa
            for variable in entity_targets_data.get("variables", []):  # noqa
                column = Variable()
                column.name = variable.get("name")
                column.dataType = variable.get("dataType")
                column.type = variable.get("type")
                entity_target_columns.append(column)
            dataset.metadata.entityAndTargets.variables = entity_target_columns
            dataset.metadata.entityAndTargets.entity = entity
            dataset.metadata.entityAndTargets.sql = entity_targets_data.get("sql")

            # Parse feature columns
            feature_columns = []  # noqa
            for variable in features_data.get("variables", []):
                column = Variable()
                column.name = variable.get("name")
                column.dataType = variable.get("dataType")
                column.type = variable.get("type")
                feature_columns.append(column)
            dataset.metadata.features.variables = feature_columns
            dataset.metadata.features.entity = entity
            dataset.metadata.features.sql = features_data.get("sql")

            # Parse predictions
            dataset.metadata.predictions.database = predictions_data.get("database")
            dataset.metadata.predictions.entity_sql = predictions_data.get("entitySql")
            dataset.metadata.predictions.table = predictions_data.get("table")

            return dataset

        except Exception as e:
            logger.error(f"Error parsing dataset: {str(e)}")
            return None

    @staticmethod
    def _datasets_to_dataframe(datasets: list[Dataset]):
        """
        Helper method to combine multiple datasets into a single DataFrame

        Parameters:
            datasets (list[Dataset]): list of datasets to combine

        Returns:
            DataFrame: Combined DataFrame containing all datasets, or None if conversion fails
        """
        if not datasets:
            return None

        try:
            from tmo.util.utils import to_dataframe

            combined_data = []
            for dataset in datasets:
                combined_data.append(dataset.get_df_template())

            return to_dataframe(combined_data)
        except Exception as e:
            logger.error(f"Could not combine datasets to DataFrame: {str(e)}")
            return None
